/*
 *  Code extracted from the retinex_pde_lib in IPOL
 *  http://www.ipol.im/pub/algo/lmps_retinex_poisson_equation/
 *
 *  Some functions added by Gloria Haro on February-June 2011.
 *
 */

#include "poisson_fft.h"


namespace libIIPStable {
	
    
    
    /**
     * @brief compute a cosinus table
     *
     * Allocate and fill a table of n values cos(i Pi / n) for i in [0..n[.
     *
     * @param size the table size
     *
     * @return the table, allocated and filled
     */
    static double *cos_table(size_t size)
    {
        double *table = NULL;
        double *ptr_table;
        size_t i;
        
        /* allocate the cosinus table */
        if (NULL == (table = (double *) malloc(sizeof(double) * size))) {
            fprintf(stderr, "allocation error\n");
            abort();
        }
        
        /*
         * fill the cosinus table,
         * table[i] = cos(i Pi / n) for i in [0..n[
         */
        ptr_table = table;
        for (i = 0; i < size; i++)
            *ptr_table++ = cos((M_PI * i) / size);
        
        return table;
    }
    
    /**
     * @brief perform a Poisson PDE in the Fourier DCT space
     *
     * @f$ u(i, j) = F(i, j) * m / (4 - 2 cos(i PI / nx)
     *                              - 2 cos(j PI / ny)) @f$
     * if @f$ (i, j) \neq (0, 0) @f$,
     * @f$ u(0, 0) = 0 @f$
     *
     * When this function is successively used on arrays of identical
     * size, the trigonometric computation is redundant and could be kept
     * in memory for a faster code. However, in our use case, the speedup
     * is marginal and we prefer to recompute this data and keep the code
     * simple.
     *
     * @param data the dct complex coefficients, of size nx x ny
     * @param nx, ny data array size
     * @param m global multiplication parameter (DCT normalization)
     *
     * @return the data array, updated
     */
    static float *poisson_dct(float *data, size_t nx, size_t ny, double m)
    {
        float *ptr_data;
        double *cosi = NULL, *cosj = NULL;
        double *ptr_cosi, *ptr_cosj;
        size_t i, j;
        double m2;
        
        /*
         * get the cosinus tables
         * cosi[i] = cos(i Pi / nx) for i in [0..nx[
         * cosj[j] = cos(j Pi / ny) for j in [0..ny[
         */
        cosi = cos_table(nx);
        cosj = cos_table(ny);
        
        /*
         * we will now multiply data[i, j] by
         * m / (4 - 2 * cosi[i] - 2 * cosj[j]))
         * and set data[i, j] to 0
         */
        m2 = - m / 2.;
        ptr_data = data;
        ptr_cosi = cosi;
        ptr_cosj = cosj;
        /*
         * handle the first value, data[0, 0] = 0
         * after that, by construction, we always have
         * *ptr_cosi + *ptr_cosj != 2.
         */
        *ptr_data++ = 0.;
        ptr_cosi++;
        /* continue with the first line from the second value */
        for (i = 1; i < nx; i++)
            *ptr_data++ *= m2 / (2. - *ptr_cosi++ - *ptr_cosj);
        ptr_cosj++;
        ptr_cosi = cosi;
        /* continue with the other lines */
        for (j = 1; j < ny; j++) {
            for (i = 0; i < nx; i++)
                *ptr_data++ *= m2 / (2. - *ptr_cosi++ - *ptr_cosj);
            ptr_cosj++;
            ptr_cosi = cosi;
        }
        
        free(cosi);
        free(cosj);
        return data;
    }
    
    void discrete_divergence(float *div, cflimage *grad_x, cflimage *grad_y, int cc)
    {
        int width = (*grad_x).w();
        int height = (*grad_x).h();
        
        float *gx_gray = grad_x[0].v(cc);
        float *gy_gray = grad_y[0].v(cc);
        float dx, dy;
        
        for (int ii=0; ii < width; ii++)
            for (int jj=0; jj < height; jj++)
            {
                int kk = jj*width + ii;
                if(ii>0) dx = gx_gray[kk] - gx_gray[kk-1];
                else dx = gx_gray[kk];
                if(ii == (width-1)) dx = - gx_gray[kk-1];
                if(jj>0) dy = gy_gray[kk] - gy_gray[kk-width];
                else dy = gy_gray[kk];
                if(jj == (height-1)) dy = - gy_gray[kk-width];
                div[kk] = dx + dy;
				
            }
    }
    
    
    void compute_mean_var(cflimage *image, float *means, float *vars)
    {
        for(int cc = 0; cc < image[0].c(); cc++)
        {
            float *igray;
            igray = image[0].v(cc);
            int counter = 0;
            
            means[cc] = 0;
            for (int kk=0; kk < image[0].wh(); kk++)
                if(igray[kk] >=0)
                {
                    means[cc] = means[cc] + igray[kk];
                    counter++;
                }
            
            means[cc] = means[cc] / counter;//image[0].wh();
            
            vars[cc] = 0;
            for (int kk=0; kk < image[0].wh(); kk++)
                if(igray[kk] >=0) vars[cc] = vars[cc] + (igray[kk] - means[cc])*(igray[kk] - means[cc]);
            
            vars[cc] = sqrt(vars[cc] / counter);//image[0].wh());
        }
    }
    
    void set_mean_var(cflimage *orig, cflimage *image, float *means_i, float *vars_i)
    {
        // Check color values
        float means_o[3], vars_o[3];
        for(int cc = 0; cc < image[0].c(); cc++)
        {
            int counter = 0;
            float *igray, *orig_gray;
            igray = image[0].v(cc);
            orig_gray = orig[0].v(cc);
            means_o[cc] = 0;
            for (int kk=0; kk < image[0].wh(); kk++)
            {
                if(orig_gray[kk] >=0)
                {
                    means_o[cc] = means_o[cc] + igray[kk];
                    counter++;
                }
                //if((igray[kk]<0) || (igray[kk]>255))
                //cout << "valor: " << igray[kk] << " canal: " << cc << endl;
            }
            means_o[cc] = means_o[cc] / counter;//image[0].wh();
            vars_o[cc] = 0;
            for(int kk=0; kk < image[0].wh(); kk++)
                if(orig_gray[kk] >=0) vars_o[cc] = vars_o[cc] + (igray[kk] - means_o[cc])*(igray[kk] - means_o[cc]);
            
            vars_o[cc] = sqrt(vars_o[cc] / counter);//image[0].wh());
        }
        
        // Set the desired mean and variance
        for(int cc = 0; cc < image[0].c(); cc++)
        {
            float *igray, *orig_gray;
            igray = image[0].v(cc);
            orig_gray = orig[0].v(cc);
            for(int kk=0; kk < image[0].wh(); kk++)
            {
                if(orig_gray[kk] >=0) igray[kk] = (igray[kk] - means_o[cc])*vars_i[cc]/vars_o[cc] + means_i[cc];
                else igray[kk] = orig_gray[kk];
            }
        }
    }
    
    void solve_Poisson_fft(cflimage *output, cflimage *grad_x, cflimage *grad_y)
    {
        fftwf_plan dct_fw, dct_bw;
        float *data_fft, *data_tmp;
        float *data = (*output).v(0);
        int nx = (*output).w(), ny = (*output).h();
        float means[3], vars[3];
        cflimage orig_image = *output;
        
        compute_mean_var(output,means,vars);
        
        /* start threaded fftw if FFTW_NTHREADS is defined */
#ifdef FFTW_NTHREADS
        if(0 == fftwf_init_threads()) {
            fprintf(stderr, "fftw initialisation error\n");
            abort();
        }
        fftwf_plan_with_nthreads(FFTW_NTHREADS);
#endif                          /* FFTW_NTHREADS */
        
        /* allocate the float-complex FFT array and the float tmp array */
        if(NULL == (data_fft = (float *) malloc(sizeof(float) * nx * ny))
           || NULL == (data_tmp = (float *) malloc(sizeof(float) * nx * ny))) {
            fprintf(stderr, "allocation error\n");
            abort();
        }
        
        
        for(int cc = 0; cc < (*output).c(); cc++)
        {
            data = (*output).v(cc);
            
            /* compute the divergence : data_tmp */
            discrete_divergence(data_tmp, grad_x, grad_y, cc);
            
            /* create the DFT forward plan and run the DCT : data_tmp -> data_fft */
            dct_fw = fftwf_plan_r2r_2d((int) ny, (int) nx,
                                       data_tmp, data_fft,
                                       FFTW_REDFT10, FFTW_REDFT10,
                                       FFTW_ESTIMATE | FFTW_DESTROY_INPUT);
            
            fftwf_execute(dct_fw);
            
            
            /* solve the Poisson PDE in Fourier space */
            /* 1. / (float) (nx * ny)) is the DCT normalisation term, see libfftw */
            poisson_dct(data_fft, nx, ny, 1. / (double) (nx * ny));
            
            /* create the DFT backward plan and run the iDCT : data_fft -> data */
            dct_bw = fftwf_plan_r2r_2d((int) ny, (int) nx,
                                       data_fft, data,
                                       FFTW_REDFT01, FFTW_REDFT01,
                                       FFTW_ESTIMATE | FFTW_DESTROY_INPUT);
            
            fftwf_execute(dct_bw);
            
            fftwf_destroy_plan(dct_fw); //
            fftwf_destroy_plan(dct_bw); //
        }
        
        set_mean_var(&orig_image,output,means,vars); 
        //compute_mean_var(output,means,vars); // sanity check
        
        /* cleanup */
        orig_image.erase();
        free(data_tmp);
        //fftwf_destroy_plan(dct_fw);
        //fftwf_destroy_plan(dct_bw);
        fftwf_free(data_fft);
        fftwf_cleanup();
#ifdef FFTW_NTHREADS
        fftwf_cleanup_threads();
#endif                          /* FFTW_NTHREADS */
    }
}

