#include "libnoiseestimation.h"



namespace libIIPStable
{
	
	
	void noise_estimate_compute_derivative(float *igray, float *igray2, int width, int height)
	{
		
		float sqrt2 = sqrtf(2.0);
		
		for(int y=0; y < height - 1; y++)
			for(int x=0; x < width - 1; x++)
			{
				
				float derivative = (float) igray[y*width+x] - (float) igray[y*width+x+1];	
				derivative /= sqrt2;
				
				igray2[y*width+x] = derivative;
				
			}
		
	}
	
	
	
	
	
	void noise_estimate_derivative_variance(float *fpI, float *fpDer, int iBlockSide,  float *fpStd, float *fpAverages, int iLengthAverages, int &iSize, int iWidth, int iHeight)
	{
		
		// size of block
		int iBlockSize = iBlockSide * iBlockSide;
		
		// for each block
		iSize = 0;
		for(int y=0; y + iBlockSide < iHeight; y++)
			for(int x=0; x + iBlockSide < iWidth; x++)
			{
				
				float dmean = 0.0;
				float dmean2 = 0.0;
				float mean = 0.0;
				
				// compute mean of the gray level
				// compute mean and square mean of derivatives
				for(int i = 0; i < iBlockSide; i++)
					for(int j = 0; j < iBlockSide; j++)
					{
						
						// gray
						float value = (float) fpI[(y+j)*iWidth + (x+i)];
						mean += value;
						
						// derivatives
						float dvalue = (float) fpDer[(y+j)*iWidth + (x+i)];
						
						dmean += dvalue;
						dmean2 += dvalue * dvalue;
						
					}
				
				// normalize averages
				mean /= (float) iBlockSize;
				dmean /= (float) iBlockSize;
				dmean2 /= (float) iBlockSize;
				
				float std = sqrtf(dmean2 - dmean*dmean);
				
				// sotre mean and std
				fpStd[iSize] = std;
				fpAverages[iSize] = mean;
				
				// increase the counter of stored values
				iSize++;
				if (iSize >= iLengthAverages)
				{
					std::cout << "warning :: " << __PRETTY_FUNCTION__ << ":: number of patches larger than storage vector length" << std::endl;
					std::cout << "returning without using all patches" << std::endl;
					return;
				}
				
			}
	}
	
	
	
	
	
	void noise_estimate_algorithm(int iBins, int iWsize, float fPercent, int iflagEqui, int iFiltIt,  float *fpLimits, float *fpX, float *fpS, float *fpI, int iWidth, int iHeight)
	{
		
		
		//// Parameters
		
		// length of the image
		int ilength = iWidth * iHeight;
		
		// correction factor applied to obtained percentile (computed empirically)
		float correction = 1.0;
		
		if (iWsize == 5 && fPercent == 0.01f) correction = 1.5907;
		else if (iWsize == 9 && fPercent == 0.01f) correction = 1.2732;
		else if (iWsize == 3 && fPercent == 0.005f) correction = 2.6933;
		else if (iWsize == 5 && fPercent == 0.005f) correction = 1.6752;
		else if (iWsize == 9 && fPercent == 0.005f) correction = 1.3084;
		else if (iWsize == 15 && fPercent == 0.01f) correction = 1.1502;
		else if (iWsize == 21 && fPercent == 0.01f) correction = 1.1035;
		else if (iWsize == 15 && fPercent == 0.005f) correction = 1.1676;
		else if (iWsize == 21 && fPercent == 0.005f) correction = 1.1153;
		else { 		printf("parameters not calibrated\n"); exit(-1);}
		
		
		//// Computing derivatives of the image
		float *fpIder = new float[ilength];
		memset(fpIder, 0.0f, ilength * sizeof(float));
		
		
		
		noise_estimate_compute_derivative(fpI, fpIder, iWidth, iHeight);
		
		
		
		///// pointers contain the gray average and variance of derivatives of all patches in the image
		float *averages = new float[ilength];
		float *stdvalues = new float[ilength];
		memset(averages, 0.0f, ilength * sizeof(float));
		memset(stdvalues, 0.0f, ilength * sizeof(float));
		
		
		///// Compute variance of patches of derivatives
		
		// Variable counting how many values are inserted in pointers averages
		int asize = 0; 
		noise_estimate_derivative_variance(fpI, fpIder,  iWsize,   stdvalues, averages, ilength, asize, iWidth, iHeight);
		
		
		
		////  Order averages and according stdvalues
		
		// stdvalues are moved at the same position of the corresponding average
		
		
		fpQuickSort(averages, stdvalues, asize);
		
		
		//// Limits of quantization steps
		memset(fpLimits, 0.0f, (iBins+1) * sizeof(float));
		
		float rMinValue = averages[0];
		float rMaxValue = averages[asize-1]; 
		
		
		int cStep;
		if (iflagEqui)
		{
			
			// histogram is partitioned
			int uasize = 0;
			for(; averages[uasize] < rMaxValue; uasize++) {;}
			
			//// Step histogram quantization
			int iStep = (int) ceilf((float) (uasize) / (float) (iBins));
			cStep = iStep;
			
			int k=0;
			for(int i = 0; i < uasize; i++)
				if (i % iStep == 0 && k < iBins)
				{
					fpLimits[k] = averages[i];
					k++;
				}
			
			fpLimits[iBins] = rMaxValue;
		}	
		else
		{
			
			// gray level range between minimum and maximum is uniformly partitiones
			float rStep = (rMaxValue - rMinValue) / (float) iBins;	
			cStep = (int) rStep;
			
			// fill fpLimits with interval extrems
			for(int i=0; i < iBins; i++)	fpLimits[i] = rMinValue + (float) i * rStep;
			fpLimits[iBins] = rMaxValue;
			
		}
		
		
		
		//// order std inside each range bin, means ordered according to std
		float * partial_order_averages = new float[asize];
		float * partial_order_stdvalues = new float[asize];
		
		for(int i=0, k=0; i < iBins; i++)
		{
			
			// Select standard deviations inside each bin 
			int counter = 0;
			while(averages[k] >= fpLimits[i] && averages[k] < fpLimits[i+1])
			{
				partial_order_averages[counter] = averages[k];
				partial_order_stdvalues[counter] = stdvalues[k];
				counter++;
				k++;
			}
			
			// Order std values. At the same time move corresponding averages
			fpQuickSort(partial_order_stdvalues, partial_order_averages, counter);
			
			// Select position of percentile
			int npercentile = (int) (fPercent * (float) counter);
			
			fpS[i] = correction * partial_order_stdvalues[npercentile];
			fpX[i] =  partial_order_averages[npercentile];
			
		}	
		
		
		
		
		
		
		if (iFiltIt > 0)
		{
			
			////////////// filter values by taking for each point the minimum between the current value 
			////////////// and the line passing between the previous and next points
			//////////////
			////////////// this is applied to the variance since in CCD's variance follows a linear model
			//////////////
			////////////// this is no longer true for JPEG images, but the strategy slightly filters the curve
			
			// take the square of std's
			for(int i=0; i < iBins; i++)
				fpS[i] *= fpS[i];
			
			
			for(int k=0; k < iFiltIt; k++)
			{
				
				for(int i=1; i < iBins - 1; i++)  
				{
					
					float x0 = fpX[i-1]; 
					float y0 = fpS[i-1];
					
					float x1 = fpX[i+1]; 
					float y1 = fpS[i+1];
					
					float py = y0 + (y1-y0)*(fpX[i]-x0) / (x1-x0);  
					
					if (fpS[i]  > py)  fpS[i] =  py;
					
				}
				
			}
			
			
			// take the square roots of var's
			for(int i=0; i < iBins; i++)
				fpS[i] = sqrtf(fpS[i]);
			
		}
		
		
		
	}
	
	
	
}
