#include "libsift.h"


namespace libIIPStable 
{
    
    
    void default_sift_parameters(siftPar &par)
    {
        par.OctaveMax=100000;
        par.DoubleImSize = 0;
        par.order = 3;
        par.InitSigma = 1.6;
        par.BorderDist = 5; 
        par.Scales = 3;
        par.PeakThresh = 255.0 * 0.04 / 3.0;
        par.EdgeThresh = 0.06; 
        par.EdgeThresh1 = 0.08; 
        par.OriBins  = 36;
        par.OriSigma = 1.5;
        par.OriHistThresh = 0.8;
        par.MaxIndexVal = 0.2;
        par.MagFactor  = 3;
        par.IndexSigma  = 1.0;
        par.IgnoreGradSign = 0;
        par.MatchRatio = 0.6;
        par.MatchXradius = 1000000.0f;
        par.MatchYradius = 1000000.0f;
        
        par.noncorrectlylocalized = 0;
        
    }

        
    
    void compute_sift_keypoints(float *input, keypointslist& keypoints, int width, int height, siftPar &par)
    {
        
        flimage image;
        
        
        // make zoom of image if necessary
        float octSize = 1.0;
        if (par.DoubleImSize){
            
            image.create(2*width, 2*height);
            apply_zoom(input,image.v(),2.0,par.order,width,height);
            octSize *= 0.5;
            
        } else
        {
            
            image.create(width,height);
            for (int i=0; i < image.wh(); i++) image[i] = input[i];
        }
        
        
        
        // suposed degree of blurring of initial image 0.5
        // if interpolated blurring 1.0
        float	curSigma;
        if (par.DoubleImSize) curSigma = 1.0; else curSigma = 0.5;
        
        
        // par.InitSigma = 1.6 (by default). Necessary to subsample.
        // convolve image to arrive to 1.6 taking into account curSigma
        if (par.InitSigma > curSigma ) {
            
            float sigma = (float) sqrt((double)(par.InitSigma * par.InitSigma - curSigma * curSigma));
            
            fiGaussianConvol(image.v(), image.v(), image.w(), image.h(), sigma);
            
        }
        
        
        
        
        // Question: why is image convolved up to 1.6 if it is convolved again before sampling
        
        
        int 	minsize = 2 * par.BorderDist + 2;	// minsize of image to be considered
        int     OctaveCounter = 0;					// number of octaves 
        
        while (image.w() > minsize &&  image.h() > minsize && OctaveCounter < par.OctaveMax) {
            
            // find key points at this octave
            OctaveKeypoints(image, octSize, keypoints,par); 
            
            
            // image is blurred inside OctaveKeypoints and therefore can be sampled by factor 2
            flimage aux( (int)((float) image.w() / 2.0f) , (int)((float) image.h() / 2.0f)); 
            
            // sample
            aux = image.subSample(2, 0);
            
            // iterate
            image = aux;
            
            // we increase octave size
            octSize *= 2.0;
            
            // we increase octave counter
            OctaveCounter++;
            
        }
        
        
        
        
        
    }
    
    
    
    /////////////////////////////////////////////////
    /// EXTREMA DETECTION IN ONE SCALE-SPACE OCTAVE:
    /////////////////////////////////////////////////
    
    
    
    
    
    
    
    // Question: par.Scales = 3 is too small??
    
    void OctaveKeypoints(flimage & image, float octSize, keypointslist& keys,siftPar &par)
    {
        
        // par.Scales: number of scales in which octave is divided
        flimage *blur = new flimage[par.Scales+3];
        flimage *dogs = new flimage[par.Scales+2];
        
        // par.Scales determine how many steps we perform to pass from one scale to the next one:  sigI --> 2*sigI
        // At each step we pass from    sigma_0 --> sigma0 * (1 + R)
        // At the last step  sigI * (1 + R)^par.Scales = 2 * sigI
        // (1+R) = 2 ^(1 / par.Scales)  it is called sigmaRatio
        
        float sigmaRatio = (float) pow(2.0, 1.0 / (double) par.Scales);
        
        // blur[0] = initial	
        blur[0] = flimage(image);	
        
        // which is convolved at par.InitSigma = 1.6 (by default)
        float prevSigma = par.InitSigma;	
        
        
        // we convolve and subsample to obtain intermediary images
        for (int i = 1; i < par.Scales + 3; i++) {
            
            blur[i] = flimage(blur[i-1]);
            
            float increase = prevSigma*(float)sqrt((double)(sigmaRatio*sigmaRatio-1.0));
            
            // we convolve by sigmaRatio the previous image
            fiGaussianConvol(blur[i].v(), blur[i].v(), blur[i].w(), blur[i].h(), increase);
            
            prevSigma *= sigmaRatio;
            
        }
        
        
        // compute an array, dogs, of difference-of-Gaussian
        for (int i = 0; i < par.Scales + 2; i++) {
            
            /// dogs[i] = blur[i] - blur[i+1]
            dogs[i] = flimage(blur[i]);		
            fpCombine(dogs[i].v(),1.0, blur[i+1].v(),-1.0, dogs[i].v(),  dogs[i].w() * dogs[i].h()); 
        }
        
        
        // Question: check if correct
        
        // image with exact blur to be subsampled is blur[scales]
        image = blur[par.Scales];
        
        // find extrema
        FindMaxMin(dogs, blur, octSize, keys,par);
        
        
        delete[] blur;
        delete[] dogs;
    }
    
    
    
    /////////////////////////////////////////////////
    ///Find the local maxima and minima of the DOG images in scale space.  Return the keypoints for these locations.
    /////////////////////////////////////////////////
    
    /// For each point at each scale we decide if it is a local maxima:
    /// - |dogs(x,y,s)| > 0.8 * par.PeakThresh
    /// - Local maximum or minimum in s-1,s,s+1
    /// - NotonEdge:  ratio of the two principle curvatures of the DOG function at this point be below a threshold.
    
    
    /// blur[par.Scales+1] is not used in order to look for extrema
    /// while these could be computed using avalaible blur and dogs
    
    
    
    
    void FindMaxMin(
                    flimage* dogs,  flimage* blur,
                    float octSize, keypointslist& keys,siftPar &par)
    {
        
        int width = dogs[0].w(), height = dogs[0].h();
        
        
        //Create an image map in which locations that have a keypoint are
        //marked with value 1.0, to prevent two keypoints being located at
        //same position.  This may seem an inefficient data structure, but
        //does not add significant overhead.
        flimage map(width,height); map= 0.0f;
        flimage grad(width,height); grad = 0.0f;
        flimage ori(width,height); ori = 0.0f;
        
        // Search through each scale, leaving 1 scale below and 1 above.
        // There are par.Scales+2 dog images.
        
        for (int s = 1; s < par.Scales+1; s++) {
            
            
            // orientation and gradient magnitude
            fiComputeImageGradient(blur[s].v(), grad.v(), ori.v(), blur[s].w(), blur[s].h(),'c');
            
            float val;
            int partialcounter = 0;
            for (int r = par.BorderDist; r < height - par.BorderDist; r++) 
                for (int c = par.BorderDist; c < width - par.BorderDist; c++) {
                    
                    // Pixel value at (c,r) position
                    val = dogs[s](c,r);	
                    
                    
                    // DOG magnitude must be above 0.8 * par.PeakThresh threshold
                    // (precise threshold check will be done once peak
                    // interpolation is performed).  Then check whether this
                    // point is a peak in 3x3 region at each level, and is not
                    // on an elongated edge.
                    if (fabs(val) > 0.8 * par.PeakThresh)
                    {
                        
                        
                        if (LocalMaxMin(val, dogs[s-1], r, c) && LocalMaxMin(val, dogs[s], r, c) && LocalMaxMin(val, dogs[s+1], r, c) && NotOnEdge(dogs[s], r, c, octSize,par))
                        {
                            
                            // increase number of keypoints
                            partialcounter++;
                            
                            // interpolate position and fill info
                            InterpKeyPoint(
                                           dogs, s, r, c, grad, ori,
                                           map, octSize, keys, 5,par);	
                            
                        } 
                        
                        
                    }
                    
                }
        }
        
    }
    
    
    // return 1 if local max in a 3x3 neighborhood
    bool LocalMax(float val, flimage& dog, int y0, int x0)
    {
        for (int x = x0 - 1; x <= x0 + 1; x++)
            for (int y = y0  - 1; y <= y0 + 1; y++){
                if (dog(x,y) > val) return 0;
            }
        
        return 1;
    }
    
    // return 1 if local min ia 3x3 neighborhood
    bool LocalMin(float val,  flimage& dog, int y0, int x0)
    {
        for (int x = x0 - 1; x <= x0 + 1; x++)
            for (int y = y0  - 1; y <= y0 + 1; y++){
                if (dog(x,y) < val) return 0;
            }
        
        return 1;
    }
    
    
    
    
    // return 1 if val < 0 and local minima 
    // return 1 if val > 0 and local maxima 
    bool LocalMaxMin(float val, const flimage& dog, int y0, int x0)
    {
        
        if (val > 0.0) {
            for (int x = x0 - 1; x <= x0 + 1; x++)
                for (int y = y0  - 1; y <= y0 + 1; y++){
                    if (dog(x,y) > val) return false;
                }
        } else {
            for (int x = x0 - 1; x <= x0 + 1; x++)
                for (int y = y0  - 1; y <= y0 + 1; y++){
                    if (dog(x,y) < val) return false;
                }
        }
        
        return true;
    }
    
    
    
    /* Returns FALSE if this point on the DOG function lies on an edge.
     This test is done early because it is very efficient and eliminates
     many points.  It requires that the ratio of the two principle
     curvatures of the DOG function at this point be below a threshold.
     
     Edge threshold is higher on the first scale where SNR is small in 
     order to reduce the number of unstable keypoints. 
     */
    int NotOnEdge(flimage& dog, int r, int c, float octSize,siftPar &par)
    {
        
        // Compute 2x2 Hessian values from pixel differences. 
        float	H00 = dog(c,r-1) - 2.0 * dog(c,r) + dog(c,r+1),
        H11 = dog(c-1,r) - 2.0 * dog(c,r) + dog(c+1,r),
        H01 = ( (dog(c+1,r+1) - dog(c-1,r+1)) - (dog(c+1,r-1) - dog(c-1,r-1)) ) / 4.0;
        
        // Compute determinant and trace of the Hessian. 
        float	det = H00 * H11 - H01 * H01,	/// Det H = \prod l_i
        trace = H00 + H11;						/// tr H = \sum l_i	
        
        /// As we do not desire edges but only corners we demand l_max / l_min less than a threshold
        /// In practice if A = k B,     A*B = k B^2
        ///				(A + B)^2 = (k+1)^2 * B^2
        ///				k B^2 >  t * (k+1)^2 * B^2 sii   k  / (k+1)^2 > t
        /// This is a decreasing function for k > 1 and value 0.3 at k=1.
        /// Setting t = 0.08, means k<=10 
        
        /* To detect an edge response, we require the ratio of smallest
         to largest principle curvatures of the DOG function
         (eigenvalues of the Hessian) to be below a threshold.  For
         efficiency, we use Harris' idea of requiring the determinant to
         be above par.EdgeThresh times the squared trace, as for eigenvalues
         A and B, det = AB, trace = A+B.  So if A = 10B, then det = 10B**2,
         and trace**2 = (11B)**2 = 121B**2, so par.EdgeThresh = 10/121 =
         0.08 to require ratio of eigenvalues less than 10.
         */
        if (octSize <= 1)
            return (det > par.EdgeThresh1 * trace * trace);
        else
            return (det > par.EdgeThresh * trace * trace);
        
    }
    
    
    /* Create a keypoint at a peak near scale space location (s,r,c), where
     s is scale (index of DOGs image), and (r,c) is (row, col) location.
     Add to the list of keys with any new keys added.
     */
    void InterpKeyPoint(
                        flimage* dogs, int s, int r, int c,
                        const flimage& grad, const flimage& ori, flimage& map,
                        float octSize, keypointslist& keys, int movesRemain,siftPar &par)
    {
        
        
        // Fit quadratic to determine offset and peak value.
        float offset[3];
        float peakval = FitQuadratic(offset, dogs, s, r, c);
        
        
        /* Move to an adjacent (row,col) location if quadratic interpolation
         is larger than 0.6 units in some direction (we use 0.6 instead of
         0.5 to avoid jumping back and forth near boundary).  We do not
         perform move to adjacent scales, as it is seldom useful and we
         do not have easy access to adjacent scale structures.  The
         movesRemain counter allows only a fixed number of moves to
         prevent possibility of infinite loops.
         */
        int newr = r, newc = c;
        if (offset[1] > 0.6 && r < dogs[0].h() - 3)
            newr++;
        else if (offset[1] < -0.6 && r > 3)
            newr--;
        
        if (offset[2] > 0.6 && c < dogs[0].w() - 3)
            newc++;
        else if (offset[2] < -0.6 && c > 3)
            newc--;
        
        if (movesRemain > 0  &&  (newr != r || newc != c)) {
            InterpKeyPoint(
                           dogs, s, newr, newc, grad, ori, map,
                           octSize, keys,movesRemain - 1,par);
            return;
        }
        
        /* Do not create a keypoint if interpolation still remains far
         outside expected limits, or if magnitude of peak value is below
         threshold (i.e., contrast is too low). */
        if (	fabs(offset[0]) > 1.5 || fabs(offset[1]) > 1.5 ||
            fabs(offset[2]) > 1.5 || fabs(peakval) < par.PeakThresh)		
        {
            par.noncorrectlylocalized++;
            return;
        }
        
        /* Check that no keypoint has been created at this location (to avoid
         duplicates).  Otherwise, mark this map location.
         */
        if (map(c,r) > 0.0) return;
        map(c,r) = 1.0;
        
        /* The scale relative to this octave is given by octScale.  The scale
         units are in terms of sigma for the smallest of the Gaussians in the
         DOG used to identify that scale.
         */
        float octScale = par.InitSigma * (float) powf(2.0f, ((float) s + offset[0]) / (float) par.Scales);
        
        /// always use histogram of orientations
        //if (UseHistogramOri)
        AssignOriHist(
                      grad, ori, octSize, octScale,
                      r + offset[1], c + offset[2], keys, par);
        //else
        //	AssignOriAvg(
        //		grad, ori, octSize, octScale,
        //		r + offset[1], c + offset[2], keys);
    }
    
    
    
    /* Apply the method developed by Matthew Brown (see BMVC 02 paper) to
     fit a 3D quadratic function through the DOG function values around
     the location (s,r,c), i.e., (scale,row,col), at which a peak has
     been detected.  Return the interpolated peak position as a vector
     in "offset", which gives offset from position (s,r,c).  The
     returned value is the interpolated DOG magnitude at this peak.
     */
    float FitQuadratic(float offset[3], flimage* dogs, int s, int r, int c)
    {
        float g[3];
        flimage *dog0, *dog1, *dog2;
        int i;
        
        //s = 1; r = 128; c = 128;
        
        laMatrix H(3, 3);
        
        /* Select the dog images at peak scale, dog1, as well as the scale
         below, dog0, and scale above, dog2. */
        dog0 = &dogs[s-1];
        dog1 = &dogs[s];
        dog2 = &dogs[s+1];
        
        /* Fill in the values of the gradient from pixel differences. */
        g[0] = ((*dog2)(c,r) - (*dog0)(c,r)) / 2.0;
        g[1] = ((*dog1)(c,r+1) - (*dog1)(c,r-1)) / 2.0;
        g[2] = ((*dog1)(c+1,r) - (*dog1)(c-1,r)) / 2.0;
        
        /* Fill in the values of the Hessian from pixel differences. */
        H[0][0] = (*dog0)(c,r)   - 2.0 * (*dog1)(c,r) + (*dog2)(c,r);
        H[1][1] = (*dog1)(c,r-1) - 2.0 * (*dog1)(c,r) + (*dog1)(c,r+1);
        H[2][2] = (*dog1)(c-1,r) - 2.0 * (*dog1)(c,r) + (*dog1)(c+1,r);
        H[0][1] = H[1][0] = ( ((*dog2)(c,r+1) - (*dog2)(c,r-1)) -
                             ((*dog0)(c,r+1) - (*dog0)(c,r-1)) ) / 4.0;
        H[0][2] = H[2][0] = ( ((*dog2)(c+1,r) - (*dog2)(c-1,r)) -
                             ((*dog0)(c+1,r) - (*dog0)(c-1,r)) ) / 4.0;
        H[1][2] = H[2][1] = ( ((*dog1)(c+1,r+1) - (*dog1)(c-1,r+1)) -
                             ((*dog1)(c+1,r-1) - (*dog1)(c-1,r-1)) ) / 4.0;
        
        /* Solve the 3x3 linear sytem, Hx = -g. Result, x, gives peak offset.
         Note that SolveLinearSystem destroys contents of H. */
        offset[0] = - g[0];
        offset[1] = - g[1];
        offset[2] = - g[2];
        
        // 	for(i=0; i < 3; i++){
        // 	
        // 		for(j=0; j < 3; j++) printf("%f  ", H[i][j]);
        // 		printf("\n");
        // 	}
        
        // 	printf("\n");
        // 
        // 	for(i=0; i < 3; i++) printf("%f  ", offset[i]);
        // 	printf("\n");
        
        laVector solution(3);
        laVector doffset(3);
        doffset[0] = offset[0];   doffset[1] = offset[1];    doffset[2] = offset[2];
        lusolve(H, solution, doffset);
        
        // 	printf("\n");
        // 	for(i=0; i < 3; i++) printf("%f  ", solution[i]);
        // 	printf("\n");
        
        
        
        
        /* Also return value of DOG at peak location using initial value plus
         0.5 times linear interpolation with gradient to peak position
         (this is correct for a quadratic approximation).
         */
        for(i=0; i < 3; i++) offset[i] = (float) solution[i];
        
        return ((*dog1)(c,r) + 0.5 * ((float) solution[0]*g[0]+ (float)solution[1]*g[1]+(float)solution[2]*g[2]));
    }
    
    
    
    
    
    
    
    
    /// - Look for local (3-neighborhood) maximum with valuer larger or equal than par.OriHistThresh * maxval
    /* Assign an orientation to this keypoint.  This is done by creating a
     Gaussian weighted histogram of the gradient directions in the
     region.  The histogram is smoothed and the largest peak selected.
     The results are in the range of -PI to PI.
     */
    
    
    // Compute histogram of orientation in a neighborhood weighted by gradient and distance to center
    void AssignOriHist(
                       const flimage& grad, const flimage& ori, float octSize,
                       float octScale, float octRow, float octCol,keypointslist& keys,siftPar &par)
    {
        int	bin, prev, next;
        float	*hist = new float[par.OriBins];
        float distsq, dif, gval, weight, angle, interp;
        float radius2, sigma2;		
        
        int	row = (int) (octRow+0.5),
        col = (int) (octCol+0.5),
        rows = grad.h(),
        cols = grad.w();
        
        for (int i = 0; i < par.OriBins; i++) hist[i] = 0.0;
        
        
        // par.OriSigma = 1.5 (by default)
        float	sigma = par.OriSigma * octScale;
        
        // neighborhood radius to look at orientation
        int	radius = (int) (sigma * 3.0);
        
        int rmin = MAX(0,row-radius);
        int cmin = MAX(0,col-radius);
        int rmax = MIN(row+radius,rows-2);
        int cmax = MIN(col+radius,cols-2);
        radius2 = (float)(radius * radius);
        sigma2 = 2.0*sigma*sigma;
        
        // look at close pixels with positive gradient and in the radius neighborhood
        // build histogram weighted by distance and magnitude of gradient
        for (int r = rmin; r <= rmax; r++) {
            for (int c = cmin; c <= cmax; c++) {
                
                gval = grad((int) c, (int) r);
                
                dif = (r - octRow);	distsq = dif*dif;
                dif = (c - octCol);	distsq += dif*dif;
                
                if (gval > 0.0  &&  distsq < radius2 + 0.5) {
                    
                    weight = exp(- distsq / sigma2);
                    
                
                    // Ori is in range of -PI to PI. 
                    // Question: is orientation in the same rang
                    
                    angle = ori((int) c, (int) r);
                    bin = (int) (par.OriBins * (angle + PI + 0.001) / (2.0 * PI));
                    assert(bin >= 0 && bin <= par.OriBins);
                    bin = MIN(bin, par.OriBins - 1);
                    hist[bin] += weight * gval;
                    
                }
                
            }
        }
        
        
        // Apply smoothing 6 times for accurate Gaussian approximation
        for (int i = 0; i < 6; i++)
            SmoothHistogram(hist, par.OriBins);
        
        
        // Find maximum value in histogram
        float maxval = 0.0;
        for (int i = 0; i < par.OriBins; i++) 
            if (hist[i] > maxval) maxval = hist[i];
        
        
        // Look for each local peak in histogram such that
        // hist[i] >= par.OriHistThresh * maximum value
        for (int i = 0; i < par.OriBins; i++) {
            
            prev = (i == 0 ? par.OriBins - 1 : i - 1);
            next = (i == par.OriBins - 1 ? 0 : i + 1);
            
            
            if (hist[i] > hist[prev]  &&  hist[i] > hist[next]  &&
                hist[i] >= par.OriHistThresh * maxval ) {
                
                // interpolate peak orientation
                interp = InterpPeak(hist[prev], hist[i], hist[next]);
                
                // Set angle in -Pi to Pi
                angle = /*(*/2.0 * PI * ((float) i + 0.5 + interp) / par.OriBins/*)*/ - PI;
                assert(angle >= -PI  &&  angle <= PI);
       
                
                // Create a keypoint with this orientation. 
                MakeKeypoint(
                             grad, ori, octSize, octScale,
                             octRow, octCol, angle, keys,par);
            }
            
        }
        
        delete[] hist;
    }
    
    
    
    /* Smooth a histogram by using a [1/3 1/3 1/3] kernel.  Assume the histogram
     is connected in a circular buffer.
     */
    void SmoothHistogram(float* hist, int bins)
    {
        float prev, temp;
        
        prev = hist[bins - 1];
        for (int i = 0; i < bins; i++) {
            temp = hist[i];
            hist[i] = ( prev + hist[i] + hist[(i + 1 == bins) ? 0 : i + 1] ) / 3.0;
            prev = temp;
        }
    }
    
    
    /* Return a number in the range [-0.5, 0.5] that represents the
     location of the peak of a parabola passing through the 3 evenly
     spaced samples.  The center value is assumed to be greater than or
     equal to the other values if positive, or less than if negative.
     */
    float InterpPeak(float a, float b, float c)
    {
        if (b < 0.0) {
            a = -a; b = -b; c = -c;
        }
        assert(b >= a  &&  b >= c);
        return 0.5 * (a - c) / (a - 2.0 * b + c);
    }
    
    
    
    
    
    /* Joan Pau: Add a new keypoint to a vector of keypoints
     Create a new keypoint and return list of keypoints with new one added.
     */
    void MakeKeypoint(
                      const flimage& grad, const flimage& ori, float octSize, float octScale,
                      float octRow, float octCol, float angle, keypointslist& keys,siftPar &par)
    {
        keypoint newkeypoint;
        newkeypoint.x = octSize * octCol;	/*x coordinate */
        newkeypoint.y = octSize * octRow;	/*y coordinate */
        newkeypoint.scale = octSize * octScale;	/* scale */
        newkeypoint.angle = angle;		/* orientation */
        MakeKeypointSample(newkeypoint,grad,ori,octScale,octRow,octCol,par);
        keys.push_back(newkeypoint);
        
        
        //printf("Adding keypoint (%f,%f,%f)\n",newkeypoint.x, newkeypoint.y, newkeypoint.angle);
        
    }
    
    
    
    /* Use the parameters of this keypoint to sample the gradient images
     at a set of locations within a circular region around the keypoint.
     The (scale,row,col) values are relative to current octave sampling.
     The resulting vector is stored in the key.
     */
    void MakeKeypointSample(
                            keypoint& key, const flimage& grad, const flimage& ori,
                            float scale, float row, float col,siftPar &par)
    {
        
        // Produce sample vector
        KeySampleVec(key, grad, ori, scale, row, col,par);
        
        
        /* Normalize vector.  This should provide illumination invariance
         for planar lambertian surfaces (except for saturation effects).
         Normalization also improves nearest-neighbor metric by
         increasing relative distance for vectors with few features.
         It is also useful to implement a distance threshold and to
         allow conversion to integer format.
         */
        
        // normalize to squared norm = 1
        NormalizeVec(key.vec);
        
        
        // threshold vector to 0.2 in order to reduce impact of a single orientation
        bool changed = false;
        for (int i = 0; i < VecLength; i++)
            if (key.vec[i] > par.MaxIndexVal) {
                key.vec[i] = par.MaxIndexVal;
                changed = true;
            }
        
        if (changed) NormalizeVec(key.vec);
        
        
        // Question: why is the vector quantized to integer
        int intval;
        for (int i = 0; i < VecLength; i++) {
            intval =  (int)(512.0 * key.vec[i]);
            key.vec[i] = (int) MIN(255, intval);
        }
        
        
    }
    
    // Normalize square norm of vector to 1.0.
    void NormalizeVec(float* vec)
    {
        float val, fac;
        
        float sqlen = 0.0;
        for (int i = 0; i < VecLength; i++) {
            val = vec[i];
            sqlen += val * val;
        }
        fac = 1.0 / sqrt(sqlen);
        
        for (int i = 0; i < VecLength; i++)
            vec[i] *= fac;
    }
    
    
    /* Create a 3D index array into which gradient values are accumulated.
     After filling array, copy values back into vec.
     */
    void KeySampleVec(
                      keypoint& key, const flimage& grad, const flimage& ori,
                      float scale, float row, float col,siftPar &par)
    {
        
        float index[IndexSize][IndexSize][OriSize];
        
        /* Initialize index array. */
        for (int i = 0; i < IndexSize; i++)
            for (int j = 0; j < IndexSize; j++)
                for (int k = 0; k < OriSize; k++)
                    index[i][j][k] = 0.0;
        
        
        KeySample(index, key, grad, ori, scale, row, col, par);
        
        
        /* Unwrap the 3D index values into 1D vec. */
        int v = 0;
        for (int i = 0; i < IndexSize; i++)
            for (int j = 0; j < IndexSize; j++)
                for (int k = 0; k < OriSize; k++)
                    key.vec[v++] = index[i][j][k];
        
    }
    
    
    
    /* Add features to vec obtained from sampling the grad and ori images
     for a particular scale.  Location of key is (scale,row,col) with respect
     to images at this scale.  We examine each pixel within a circular
     region containing the keypoint, and distribute the gradient for that
     pixel into the appropriate bins of the index array.
     */
    void KeySample(
                   float index[IndexSize][IndexSize][OriSize], keypoint& key,
                   const flimage& grad, const flimage& ori, float scale, float row, float col,siftPar &par)
    {
        float rpos, cpos, rx, cx;
        
        int	irow = (int) (row + 0.5),
        icol = (int) (col + 0.5);
        float	sine   = (float) sin(key.angle),
        cosine = (float) cos(key.angle);
        
        /* The spacing of index samples in terms of pixels at this scale. */
        float	spacing = scale * (float) par.MagFactor;
        
        /* Radius of index sample region must extend to diagonal corner of
         index patch plus half sample for interpolation. */
        float	radius = 1.414 * spacing * (IndexSize + 1) / 2.0;
        int	iradius = (int) (radius + 0.5);
        
        /* Examine all points from the gradient image that could lie within the
         index square. */
        for (int i = -iradius; i <= iradius; i++) {
            for (int j = -iradius; j <= iradius; j++) {
                
                /* Rotate sample offset to make it relative to key orientation.
                 Uses (row,col) instead of (x,y) coords.  Also, make subpixel
                 correction as later image offset must be an integer.  Divide
                 by spacing to put in index units.
                 */
            

                /* Guoshen Yu, inverse the rotation */
                rpos = ((cosine * (float) i - sine * (float) j) - (row - (float) irow)) / spacing;
                cpos = ((sine * (float) i + cosine * (float) j) - (col - (float) icol)) / spacing;
                
                /*
                 rpos = ((cosine * i + sine * j) - (row - irow)) / spacing;
                 cpos = ((- sine * i + cosine * j) - (col - icol)) / spacing;*/
                
                /* Compute location of sample in terms of real-valued index array
                 coordinates.  Subtract 0.5 so that rx of 1.0 means to put full
                 weight on index[1] (e.g., when rpos is 0 and IndexSize is 3. */
                rx = rpos + (float) IndexSize / 2.0 - 0.5;
                cx = cpos + (float) IndexSize / 2.0 - 0.5;
                
                /* Test whether this sample falls within boundary of index patch. */
                if (	rx > -1.0 && rx < (float) IndexSize  &&
                    cx > -1.0 && cx < (float) IndexSize )
                    AddSample(
                              index, key, grad, ori,
                              (float) (irow + i), (float) (icol + j), rpos, cpos, rx, cx,par);
            }
        }
    }
    
    
    
    
    
    /* Given a sample from the image gradient, place it in the index array.
     */
    void AddSample(
                   float index[IndexSize][IndexSize][OriSize], keypoint& key,
                   const flimage& grad, const flimage& orim,
                   float r, float c, float rpos, float cpos, float rx, float cx,siftPar &par)
    {
        /* Clip at image boundaries. */
        if (r < 0  ||  r >= grad.h()  ||  c < 0  ||  c >= grad.w())
            return;
        
        /* Compute Gaussian weight for sample, as function of radial distance
         from center.  Sigma is relative to half-width of index. */
        float	sigma  = par.IndexSigma * 0.5 * IndexSize,
        weight = exp(- (rpos * rpos + cpos * cpos) / (2.0 * sigma * sigma)),
        mag    = weight * grad((int) c, (int) r);
        
        /* Subtract keypoint orientation to give ori relative to keypoint. */
        float	ori = orim((int) c, (int) r) - key.angle;
        
        /* Put orientation in range [0, 2*PI].  If sign of gradient is to
         be ignored, then put in range [0, PI]. */
       
        if (par.IgnoreGradSign) 
        {
            while (ori > PI ) ori -= PI;
            while (ori < 0.0) ori += PI;
        
        } else 
        {
            
            while (ori > 2.0*PI) ori -= 2.0*PI;
            while (ori < 0.0   ) ori += 2.0*PI;
            
            
            // A.Buades 21/03/2012
            /*
            if (ori > PI) {ori -= PI; ori = -1.0*ori;}
            else if (ori < -PI) { ori += PI; ori = -1.0*ori;}
            ori += PI;
            */
        }
        
        
        PlaceInIndex(index, mag, ori, rx, cx,par);
    }
    
   
    
    

    
    
    
    
    
    
    
    
    /* Increment the appropriate locations in the index to incorporate
     this image sample.  The location of the sample in the index is (rx,cx).
     */
    void PlaceInIndex(
                      float index[IndexSize][IndexSize][OriSize],
                      float mag, float ori, float rx, float cx,siftPar &par)
    {
        int	orr, rindex, cindex, oindex;
        float	rweight, cweight, oweight;
        float  *ivec;
        
        float	oval = (float) OriSize * ori / (par.IgnoreGradSign ? PI : 2.0*PI);
        
        int ri = (int) ((rx >= 0.0) ? rx : rx - 1.0);	/* Round down to next integer. */
        int ci = (int) ((cx >= 0.0) ? cx : cx - 1.0);
        int oi = (int) ((oval >= 0.0) ? oval : oval - 1.0);
	//int ri = floor(rx);
	//int ci = floor(rx);
	//int oi = floor(rx);
        float rfrac = rx - ri;			/* Fractional part of location. */
        float cfrac = cx - ci;
        float ofrac = oval - oi;
	if (ri >= IndexSize) ri = IndexSize - 1;
	//fprintf(stderr, "rx => ri  :  %g => %d\n", rx, ri);
	if (ri < -1) ri = -1;
        assert(ri >= -1);
      	assert(ri < IndexSize);
	assert(oi >=  0);
	assert(oi <= OriSize);
	assert(rfrac >= 0.0);
	assert(rfrac <= 1.0);
        
        /* Put appropriate fraction in each of 8 buckets around this point
         in the (row,col,ori) dimensions.  This loop is written for
         efficiency, as it is the inner loop of key sampling. */
        for (int r = 0; r < 2; r++) {
            rindex = ri + r;
            if (rindex >=0 && rindex < IndexSize) {
                rweight = mag * ((r == 0) ? 1.0 - rfrac : rfrac);
                
                for (int c = 0; c < 2; c++) {
                    cindex = ci + c;
                    if (cindex >=0 && cindex < IndexSize) {
                        cweight = rweight * ((c == 0) ? 1.0 - cfrac : cfrac);
                        ivec = index[rindex][cindex];
                        for (orr = 0; orr < 2; orr++) {
                            oindex = oi + orr;
                            if (oindex >= OriSize)  /* Orientation wraps around at PI. */
                                oindex = 0;
                            oweight = cweight * ((orr == 0) ? 1.0 - ofrac : ofrac);
                            ivec[oindex] += oweight;
                        }
                    }
                }
            }
        }
    }
   
    
    
     
    
    
    
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /// SIFT keypoint matching 
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    float DistSquared(keypoint &k1,keypoint &k2, float tdist, siftPar &par)
    {
        
        float dif;
        float distsq = 0.0;
        
        if (fabs(k1.x - k2.x) > par.MatchXradius || fabs(k1.y - k2.y) > par.MatchYradius) return tdist;
        
        float *ik1 = k1.vec;
        float *ik2 = k2.vec;
        
        for (int i = 0; i < VecLength && distsq <= tdist; i++) 
        {
            dif = ik1[i] - ik2[i];
            distsq += dif * dif;
        }
        
        return distsq;
    }
    
    
    
    
    /* This searches through the keypoints in klist for the two closest
     matches to key.  It returns the ratio of the distance to key of the
     closest and next to closest keypoints in klist, while bestindex is the index
     of the closest keypoint.
     */
    float CheckForMatch(
                        keypoint& key, keypointslist& klist, int& min,siftPar &par)
    {
        
        int	nexttomin = -1;
        float	dsq, distsq1, distsq2;
        distsq1 = distsq2 = 1000000000000.0f;
        
        for (int j=0; j< (int) klist.size(); j++){
            
            dsq = DistSquared(key, klist[j], distsq2,par);
            
            if (dsq < distsq1) {
                distsq2 = distsq1;
                distsq1 = dsq;
                nexttomin = min;
                min = j;
            } else if (dsq < distsq2) {
                distsq2 = dsq;
                nexttomin = j;
            }
        }
        
        return distsq1/distsq2 ;
    }
    
    
    void compute_sift_matches(keypointslist& keys1,  keypointslist& keys2,
                              matchingslist& matchings,siftPar &par)
    {
        int	imatch = 0;
        float	sqminratio = par.MatchRatio * par.MatchRatio,
        sqratio;
        
        for (int i=0; i< (int) keys1.size(); i++) {
            
            sqratio = CheckForMatch(keys1[i], keys2, imatch, par);
            
            if (sqratio< sqminratio)
                matchings.push_back( matching(keys1[i],keys2[imatch] ));
        }
    }
    
    
    
    
    
}
