// Authors: Unknown. Please, if you are the author of this file, or if you 
// know who are the authors of this file, let us know, so we can give the 
// adequate credits and/or get the adequate authorizations.

// WARNING: 
// This file implements an algorithm possibly linked to the patent
//
// David Lowe  "Method and apparatus for identifying scale invariant 
// features in an image and use of same for locating an object in an 
// image",  U.S. Patent 6,711,293.
//
// This file is made available for the exclusive aim of serving as
// scientific tool to verify of the soundness and
// completeness of the algorithm description. Compilation,
// execution and redistribution of this file may violate exclusive
// patents rights in certain countries.
// The situation being different for every country and changing
// over time, it is your responsibility to determine which patent
// rights restrictions apply to you before you compile, use,
// modify, or redistribute this file. A patent lawyer is qualified
// to make this determination.
// If and only if they don't conflict with any patent terms, you
// can benefit from the following license terms attached to this
// file.
//
// This program is provided for scientific and educational only:
// you can use and/or modify it for these purposes, but you are
// not allowed to redistribute this work or derivative works in
// source or executable form. A license must be obtained from the
// patent right holders for any other use.


#include "libsiftIPOL.h"





namespace libSiftIPOLCODE
{
    
    
    
#define DEBUG 0
    
#define ABS(x)    (((x) > 0) ? (x) : (-(x)))
    
    
    
    
    
    
    //////////////////////////////////////////////// Class flimage
    //// Construction
    flimage::flimage() : width(0), height(0), p(0) 
    {
    }	 
    
    flimage::flimage(int w, int h) : width(w), height(h), p(new float[w*h]) 
    {
        for (int j=width*height-1; j>=0 ; j--) p[j] = 0.0;
    }	
    
    
    flimage::flimage(int w, int h, float v) : width(w), height(h), p(new float[w*h]) 
    {
        for (int j=width*height-1; j>=0 ; j--) p[j] = v;
    }
    
    
    flimage::flimage(int w, int h, float* v) : width(w), height(h), p(new float[w*h]) 
    {
        for (int j=width*height-1; j>=0 ; j--) p[j] = v[j];
    }
    
    
    void flimage::create(int w, int h)
    { 
        erase();
        width = w; height = h; 
        p = new float[w*h];	
        for (int j=width*height-1; j>=0 ; j--) p[j] = 0.0;
    }
    
    void flimage::create(int w, int h, float* v)
    {
        erase();
        width = w; height = h;  p = new float[w*h]; 
        for (int j=width*height-1; j>=0 ; j--) p[j] = v[j];
    }
    
    
    flimage::flimage(const flimage& im) : width(im.width), height(im.height), p(new float[im.width*im.height]) 
    {
        for (int j=width*height-1; j>=0 ; j--) p[j] = im.p[j];
    }
    
    flimage&  flimage::operator= (const flimage& im)
    {	
        if (&im == this) {
            return *this;
        }
        
        if (width != im.width || height != im.height)
        {  			
            erase();
            width = im.width; height=im.height; p = new float[width*height];
        }
        
        for (int j=width*height-1; j>=0 ; j--) p[j] = im.p[j];
        return *this;	
    }
    
    
    //// Destruction
    void flimage::erase() 
    {
        width = height = 0; 
        if (p) delete[] p; 
        p=0;
    } 
    
    flimage::~flimage()
    {
        erase();
    }
    
    
    
    
    
    
    void wxwarning(const char * message, const char *function,const char *file)
    {
        
        printf("warning :: %s :: %s :: %s\n", file, function, message);
        
    }
    
    
    void wxerror(const char * message, const char *function, const char *file)
    {
        
        printf("error :: %s :: %s :: %s\n", file, function, message);
        exit(-1);
        
    }
    
    
    
    
    double fsqr(double a) {	return a*a; }
    
    
    void  fill_exp_lut(float *lut, int size)
    {
        for(int i=0; i< size;i++)   lut[i]=expf( - (float) i / LUTPRECISION);
    }
    
    
    
    
    /* Looks for f(dif) in the lut */
    float slut(float dif,float *lut)
    {
        if (dif >= (float) LUTMAX) return 0.0;
        
        int  x= (int) floor(dif*LUTPRECISION);
        
        float y1=lut[x];
        float y2=lut[x+1];
        
        return y1 + (y2-y1)*(dif*LUTPRECISION - (float) x); 
    }
    
    
    
    
    float max(float *u,int *pos, int size)
    {  
        float max=u[0];
        if (pos) *pos=0;  
        for(int i=1; i<size; i++)  if(u[i]>max){ max=u[i]; if (pos) *pos=i; } 
        return max;
    }
    
    
    float min(float *u,int *pos,int size) 
    {
        float min=u[0];  
        if (pos) *pos=0;
        for(int i=1;i<size;i++)	if(u[i]<min){ min=u[i]; if (pos) *pos=i; }
        return min;
    }
    
    
    void max_u_v(float *u,float *v,int size) { for(int i=0;i<size;i++)  u[i]=MAX(u[i],v[i]);}
    
    void max_u_k(float *u, float k, int size) { for(int i=0;i<size;i++)  u[i]=MAX(u[i],k);}
    
    void min_u_v(float *u, float *v,int size) { for(int i=0;i<size;i++)  u[i]=MIN(u[i],v[i]);}
    
    void min_u_k(float *u,float k,int size) { int i=0; for(i=0;i<size;i++)  u[i]=MIN(u[i],k);}
    
    
    
    void copy(float *u,float *v,int size) {  for(int i=0; i<size ;i++)  v[i]=u[i]; }
    
    void clear(float *u,float value,int size) { for(int i=0; i<size; i++) u[i]=value;  }
    
    void combine(float *u,float a,float *v,float b, float *w,  int size)  { for(int i=0;i<size ;i++)   w[i]= a*u[i] + b*v[i];  }
    
    void multiple(float *u,float multiplier,int size) { for(int i=0;i<size;i++)  u[i]=multiplier*u[i]; }
    
    float scalar_product(float *u, float *v, int n)
    {
        
        float aux = 0.0f;
        for(int i=0; i < n; i++)
            aux += u[i] * v[i];
        
        return aux;
        
    }
    
    
    
    float  var(float *u,int size)
    {
        
        float *ptru=&u[0];
        float mean=0.0;
        float mean2 = 0.0;
        for(int i=0;i<size;i++,ptru++) { mean +=  *ptru; mean2 += *ptru * (*ptru);}
        
        mean/=(float) size; mean2/=(float) size;
        float var = mean2- mean*mean;
        
        return var;
    }
    
    
    
    float  mean(float *u,int size)
    {
        
        float *ptru=&u[0];
        float mean=0.0;
        for(int i=0; i<size; i++,ptru++)  mean +=  *ptru;
        mean/=(float) size;
        return mean;
    }
    
    
    
    float  median(float *u,int size)
    {
        
        float *vector = new float[size];
        float *index = new float[size];
        
        
        float *ptru=&u[0];
        for(int i=0; i<size; i++,ptru++)  { vector[i] = *ptru; index[i] = (float) i;}
        quick_sort(vector,index,size);	
        
        float median;
        if (size%2==1)	median = vector[size/2];
        else	median = (vector[size/2] + vector[(size/2) - 1])/2;
        
        delete[] vector; delete[] index;
        return median;
    }
    
    
    
    
    int normalize(float *u,int size) 
    {
        
        float some = 0.0;  
        for(int i=0;i<size;i++) some+=u[i];
        if (some != 0.0) {	for(int i=0;i<size;i++) u[i]/=some;  return 1;}  
        else return 0;	
    }
    
    
    
    float  nearest(float *u,float value,int *pos,int size)
    {
        float mindif =  fabsf(value - u[0]);
        float minvalue = u[0];
        if (pos) *pos = 0;
        
        for(int i=0;i<size;i++){
            
            float dif = fabsf(value - u[i]);
            if (dif < mindif) { mindif=dif; minvalue=u[i]; if (pos!=NULL) *pos=i;} 
            
        }
        
        return minvalue;
    }
    
    
    void binarize(float *u, float *v,float value, int inverse, int size)
    { 
        for(int i=0;i<size;i++){
            
            if (u[i] >= value && !inverse) 	v[i]= 255.0;
            else if (u[i] <= value && inverse)  v[i]= 255.0;
            else v[i]= 0.0;
        }
    }
    
    
    
    float *  gauss(int sflag,float std,int *size)
    {
        
        float *u,prec = 4.0,shift;
        double v;
        //   int n,i,flag;
        int n,i;
        int flag = 1; //Guoshen Yu
        
        
        if (sflag) n=*size;
        else
            n = 1+2*(int)ceil((double)std*sqrt(prec*2.*log(10.)));   
        
        u =new float[n];
        
        if (n==1) 
            u[0]=1.0;
        else{
            
            shift = 0.5*(float)(n-1);
            
            for (i=(n+1)/2;i--;) {
                v = ((double)i - (double) shift)/(double)std;
                u[i] = u[n-1-i] = (float) exp(-0.5*v*v); 
            }
        }	
        
        //   if (flag = normalize(u,n)) {
        if (flag == normalize(u,n)) 
        {
            *size=n;
            return u;
        } 
        else 
        {
            printf("ERROR: _gauss: _normalize: normalization equals zero.\n");
            delete[] u; /*memcheck*/
            return 0; // Guoshen Yu
        }
    }
    
    
    
    /* Quicksort,  values in arr are set in increasing order and brr elements are switched at the same time*/
    void FSWAP(float *x,float *y)
    {
        float aux;
        aux=*x;
        *x=*y;
        *y=aux;
    }
    
    
    
    void quick_sort(float *arr,float *brr,int n)
    {
        int M=7,NSTACK=50;
        int i,ir,j,k,jstack=-1,l=0;
        float a,b;
        int istack[50];
        
        ir=n-1;
        
        
        for(;;){
            if(ir-l<M){
                for(j=l+1;j<=ir;j++){
                    a=arr[j];
                    b=brr[j];
                    for(i=j-1;i>=l;i--){
                        if (arr[i]<=a) break;
                        arr[i+1]=arr[i];
                        brr[i+1]=brr[i];
                    }
                    arr[i+1]=a;
                    brr[i+1]=b;
                    
                }
                
                if (jstack<0) break;
                ir=istack[jstack--];
                l=istack[jstack--];
            } else {
                
                k=(l+ir) >> 1;
                FSWAP(&arr[k],&arr[l+1]);
                FSWAP(&brr[k],&brr[l+1]);
                if (arr[l]>arr[ir]){
                    FSWAP(&arr[l],&arr[ir]);
                    FSWAP(&brr[l],&brr[ir]);
                }
                if (arr[l+1]>arr[ir]){
                    FSWAP(&arr[l+1],&arr[ir]);
                    FSWAP(&brr[l+1],&brr[ir]);
                }
                if (arr[l]>arr[l+1]){
                    FSWAP(&arr[l],&arr[l+1]);
                    FSWAP(&brr[l],&brr[l+1]);
                }
                i=l+1;
                j=ir;
                a=arr[l+1];
                b=brr[l+1];
                for(;;){
                    do i++; while (arr[i]<a);
                    do j--; while (arr[j]>a);
                    if (j<i) break;
                    FSWAP(&arr[i],&arr[j]);
                    FSWAP(&brr[i],&brr[j]);
                }
                
                arr[l+1]=arr[j];
                arr[j]=a;
                brr[l+1]=brr[j];
                brr[j]=b;
                jstack+=2;
                
                if (jstack>=NSTACK) { printf("Stack too small\n"); exit(-1);}
                if (ir-i+1>=j-l){
                    istack[jstack]=ir;
                    istack[jstack-1]=i;
                    ir=j-1;
                } else {
                    istack[jstack]=j-1;
                    istack[jstack-1]=l;
                    l=i;
                }
                
            }
        }
        
        
    }
    
    
    
    /// histogram of values. 'n' (number of bins) or  's' (step) must be selected in flag while the other value is filled 
    float * histo(float* input, float *iminim, float *imaxim, int *n, float *s, int size, char flag)
    {
        
        if (flag != 's' && flag != 'n') 	{ printf("Warning (histo): Please select s or n as flag\n");  return NULL;	}
        
        float minim;
        if (iminim) minim = *iminim;
        else minim = min(input, NULL, size);
        
        float maxim;
        if (imaxim) maxim = *imaxim;
        else maxim = max(input, NULL, size);
        
        int num;
        float step;
        if (flag == 'n')
        {
            num = *n;
            step = (maxim-minim)/ (float)num;
            *s = step;
        } else
        {
            step = *s;
            num = (int)(0.5+(maxim-minim)/step);
            *n = num;
        }
        
        float *histo = new float[num];
        clear(histo,0.0,num);
        
        
        for(int i=0; i < size; i++)
        {
            
            int cell = (int) floorf((input[i]-minim) / step);
            
            if (cell < 0) cell = 0;
            if (cell >= num) cell = num - 1;
            
            histo[cell]++;
        }
        
        
        return histo;
        
    }
    
    
    
    
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////   Used and checked image functions
    
    void compute_gradient_orientation(float* igray,float *grad, float *ori, int width, int height)
    {
        
        float xgrad, ygrad;
        int rows, cols, r, c;
        
        rows = height;
        cols = width;
        
        
        for (r = 0; r < rows; r++)
            for (c = 0; c < cols; c++) {
                if (c == 0)
                    xgrad = 2.0 * (igray[r*cols+c+1] - igray[r*cols+c]);
                else if (c == cols-1)
                    xgrad = 2.0 * (igray[r*cols+c] - igray[r*cols+c-1]);
                else
                    xgrad = igray[r*cols+c+1] - igray[r*cols+c-1];
                if (r == 0)
                    ygrad = 2.0 * (igray[r*cols+c] - igray[(r+1)*cols+c]);
                else if (r == rows-1)
                    ygrad = 2.0 * (igray[(r-1)*cols+c] - igray[r*cols+c]);
                else
                    ygrad = igray[(r-1)*cols+c] - igray[(r+1)*cols+c];
                
                
                if (grad) grad[r*cols+c] = (float)sqrt((double)(xgrad * xgrad + ygrad * ygrad));
                if (ori) ori[r*cols+c] = (float)atan2 (-(double)ygrad,(double)xgrad);
                
            }
    }
    
    
    
    
    void sample(float *igray,float *ogray, float factor, int width, int height)
    {
        
        int swidth = (int)((float) width / factor);
        int sheight = (int)((float) height / factor);
        
        for(int j=0; j < sheight; j++)
            for(int i=0; i < swidth; i++)
                ogray[j*swidth + i] = igray[(int)((float) j * factor) * width + (int) ((float) i*factor)];
        
    }
    
    
    void sample_aglomeration(float *igray,float *ogray, float factor, int width, int height)
    {
        
        int swidth = (int)((float) width / factor);
        int sheight = (int)((float) height / factor);
        int ssize = swidth * sheight;
        
        clear(ogray,0.0,swidth*sheight);
        
        for(int j=0; j < height; j++)
            for(int i=0; i < width; i++){
                int index = (int)((float) j / factor) * swidth + (int) ((float) i / factor);
                if (index < ssize) ogray[index]  += igray[j*width+i];
            }
        
        factor *= factor;
        for(int i = 0; i < swidth*sheight; i++)
            ogray[i] /= factor; 
        
        
    }
    
    
    /*
     void extract(float *igray,float *ogray, int ax, int ay,int cwidth, int cheight,int width, int height)
     {
     for(int j=0; j < cheight; j++)
     for(int i=0; i < cwidth; i++)	
     ogray[j*cwidth + i] = igray[(ay+j)*width + ax+i];	
     }
     */
    
    
    void gray(float *red, float *green,float *blue, float *out, int width, int height)
    {
        for(int i=width*height-1; i>0; i--)	out[i] = (red[i] + green[i] + blue[i]) /3.0;
    }
    
    
    /*
     float l2_distance(float *u0,float *u1,int i0,int j0,int i1,int j1,int radius,int width,int height)
     {
     
     int wsize=(2*radius+1)*(2*radius+1);
     
     float dist=0.0;       
     for (int s=-radius; s<= radius; s++){
     
     int l = (j0+s)*width + (i0-radius);
     float *ptr0 = &u0[l];
     
     l = (j1+s)*width + (i1-radius);
     float *ptr1 = &u1[l];
     
     for(int r=-radius;r<=radius;r++,ptr0++,ptr1++){	float dif = (*ptr0 - *ptr1); dist += (dif*dif); }
     
     }
     
     dist/=(float) wsize;
     return dist;
     }
     
     
     float l2_distance_non_normalized(float *u0,float *u1,int i0,int j0,int i1,int j1,int radius,int width,int height)
     {
     
     
     float dist=0.0;       
     for (int s=-radius; s<= radius; s++){
     
     int l = (j0+s)*width + (i0-radius);
     float *ptr0 = &u0[l];
     
     l = (j1+s)*width + (i1-radius);
     float *ptr1 = &u1[l];
     
     for(int r=-radius;r<=radius;r++,ptr0++,ptr1++){	float dif = (*ptr0 - *ptr1); dist += (dif*dif); }
     
     }
     
     return dist;
     }
     
     
     float l2_distance_nsq(float *u0,float *u1,int i0,int j0,int i1,int j1,int xradius, int yradius,int width,int height)
     {
     
     int wsize=(2*xradius+1)*(2*yradius+1);
     
     float dist=0.0;       
     for (int s=-yradius; s <= yradius; s++){
     
     int l = (j0+s)*width + (i0-xradius);
     float *ptr0 = &u0[l];
     
     l = (j1+s)*width + (i1-xradius);
     float *ptr1 = &u1[l];
     
     for(int r=-xradius;r<=xradius;r++,ptr0++,ptr1++)
     {
     float dif = (*ptr0 - *ptr1); dist += (dif*dif); 
     }
     
     }
     
     dist/=(float) wsize;
     return dist;
     }
     
     
     
     
     float weighted_l2_distance(float *u0,float *u1,int i0,int j0,int i1,int j1,int width,int height,float *kernel,int radius)
     {
     
     
     float *ptrk=&kernel[0];
     float dist=0.0;       
     for (int s=-radius; s<= radius; s++){
     
     int l = (j0+s)*width + (i0-radius);
     float *ptr0 = &u0[l];
     
     l = (j1+s)*width + (i1-radius);
     float *ptr1 = &u1[l];
     
     
     for(int r=-radius;r<=radius;r++,ptr0++,ptr1++,ptrk++){ float dif = (*ptr0 - *ptr1); dist += *ptrk*(dif*dif); }
     
     }
     
     return dist;
     }
     
     
     
     float weighted_l2_distance_nsq(float *u0,float *u1,int i0,int j0,int i1,int j1,int width,int height,float *kernel,int xradius, int yradius)
     {
     
     
     float *ptrk=&kernel[0];
     float dist=0.0;       
     for (int s=-yradius; s<= yradius; s++){
     
     int l = (j0+s)*width + (i0-xradius);
     float *ptr0 = &u0[l];
     
     l = (j1+s)*width + (i1-xradius);
     float *ptr1 = &u1[l];
     
     
     for(int r=-xradius;r<=xradius;r++,ptr0++,ptr1++,ptrk++){ float dif = (*ptr0 - *ptr1); dist += *ptrk*(dif*dif); }
     
     }
     
     return dist;
     }
     */
    
    
    
    /// RGV to YUV conversion
    void rgb2yuv(float *r,float *g,float *b,float *y,float *u,float *v,int width,int height) 
    {
        int size=height*width;
        
        for(int i=0;i<size;i++){
            y[i] = COEFF_YR * r[i] + COEFF_YG * g[i] + COEFF_YB * b[i];
            u[i] = r[i] - y[i];
            v[i] = b[i] - y[i];
        }
    }
    
    
    
    void rgb2yuv(float *r,float *g,float *b,float *y,float *u,float *v,float yR, float yG, float yB, int width,int height) 
    {
        int size=height*width;
        
        
        for(int i=0;i<size;i++){
            y[i] = yR * r[i] + yG * g[i] + yB * b[i];
            u[i] = r[i] - y[i];
            v[i] = b[i] - y[i];
        }
    }
    
    
    
    /// YUV to RGB conversion
    void yuv2rgb(float *r,float *g,float *b,float *y,float *u,float *v,int width,int height)  
    {
        int size=height*width;
        
        for(int i=0;i<size;i++){
            g[i] = (y[i] - COEFF_YR * (u[i] + y[i]) - COEFF_YB * (v[i] + y[i])) / COEFF_YG;
            r[i] = u[i] + y[i];
            b[i] = v[i] + y[i];
        }
    }
    
    
    void yuv2rgb(float *r,float *g,float *b,float *y,float *u,float *v,float yR, float yG, float yB, int width,int height)  
    {
        int size=height*width;
        
        for(int i=0;i<size;i++){
            g[i] = (y[i] - yR * (u[i] + y[i]) - yB * (v[i] + y[i])) / yG;
            r[i] = u[i] + y[i];
            b[i] = v[i] + y[i];
        }
    }
    
    
    void draw_line(float *igray, int a0, int b0, int a1, int b1, float value, int width, int height)
    {
        
        int bdx,bdy;
        int sx,sy,dx,dy,x,y,z,l;
        
        bdx = width;
        bdy = height;
        
        if (a0 < 0) a0=0; 
        else if (a0>=bdx) a0=bdx-1;
        
        if (a1<0)  a1=0; 
        else  if (a1>=bdx)   a1=bdx-1;
        
        if (b0<0) b0=0; 
        else if (b0>=bdy) b0=bdy-1;
        
        if (b1<0) 	b1=0; 
        else if (b1>=bdy) b1=bdy-1; 
        
        if (a0<a1) { sx = 1; dx = a1-a0; } else { sx = -1; dx = a0-a1; }
        if (b0<b1) { sy = 1; dy = b1-b0; } else { sy = -1; dy = b0-b1; }
        x=0; y=0;
        
        if (dx>=dy) 
        {
            z = (-dx) / 2;
            while (abs(x) <= dx) 
            {
                
                l =  (y+b0)*bdx+x+a0; 
                
                igray[l] = value;
                
                x+=sx;
                z+=dy;
                if (z>0) { y+=sy; z-=dx; }
                
            } 
            
        }
        else 
        {
            z = (-dy) / 2;
            while (abs(y) <= dy) {
                
                l = (y+b0)*bdx+x+a0;
                igray[l] = value;
                
                y+=sy;
                z+=dx;
                if (z>0) { x+=sx; z-=dy; }
            }
        }
        
    }
    
    
    
    void draw_square(float *igray, int a0, int b0, int w0, int h0, float value, int width, int height)
    {
        
        draw_line(igray,a0,b0,a0+w0,b0,value,width,height);
        draw_line(igray,a0,b0,a0,b0+h0,value,width,height);
        draw_line(igray,a0+w0,b0,a0+w0,b0+h0,value,width,height);
        draw_line(igray,a0,b0+h0,a0+w0,b0+h0,value,width,height);
        
    }
    
    
    
    
    
    
    /////////////////////////////////////////////////////////////// Build Gaussian filters
    float * directional_gauss_filter(float xsigma, float ysigma, float angle, int *kwidth, int *kheight)
    {
        
        
        int ksize = (int)(2.0 * 2.0 * MAX(xsigma, ysigma) + 1.0);
        float *kernel = new float[ksize*ksize];
        
        float xsigma2 = xsigma*xsigma;
        float ysigma2 = ysigma*ysigma;
        
        int l2 = ksize/2;
        for(int y = -l2; y <= l2; y++)
            for(int x = -l2; x <= l2; x++)
            {
                
                float a = (float) angle * PI / 180.0f;
                float sina = sin(a);
                float cosa = cos(a);
                
                float ax = (float) x * cosa + (float) y * sina;
                float ay = -(float) x * sina + (float) y * cosa;
                kernel[(y+l2) * ksize + x + l2] =  exp(-(ax*ax)/(2.0f*xsigma2)  - (ay*ay)/(2.0f*ysigma2) );  
                
            }
        
        
        float sum=0.0;
        for(int i=0; i < ksize*ksize; i++) sum += kernel[i];
        for(int i=0; i < ksize*ksize; i++) kernel[i] /= sum;
        
        *kwidth = ksize;
        *kheight = ksize;
        
        return kernel;
    }
    
    
    
    
    
    
    /* Convolution with a kernel */
    /* No padding applied to the image */
    void convol(float *u,float *v,int width,int height,float *kernel,int kwidth,int kheight)
    {
        
        //  float S;
        //  int K2,L2,m,n,kmin,kmax,lmin,lmax,l,k;
        
        int K2 = kwidth / 2;
        int L2 = kheight / 2;
        
        for(int y=0 ; y < height; y++) 
            for (int x=0 ; x < width; x++) {
                
                float S = 0.0;
                //      kmax = MIN(kwidth-1,n+K2);
                //      kmin = MAX(0,1+n+K2-width);
                //      lmax = MIN(kheight-1,m+L2);
                //      lmin = MAX(0,1+m+L2-height);
                
                for (int l = -L2; l <= L2; l++) 
                    for (int k = -K2 ; k<= K2; k++)
                    { 
                        int px=x+k;
                        int py=y+l;
                        
                        if (px>=0 && px < width && py>=0 && py<height)
                            S += u[width*py + px] * kernel[kwidth*(l+L2) + k+K2];
                    }
                
                v[y*width+x] = (float) S;
                
            }
    }
    
    
    
    void median(float *u,float *v, float radius, int niter, int width,int height)
    {
        
        
        int iradius = (int)(radius+1.0);
        int rsize=(2*iradius+1)*(2*iradius+1);
        
        float * vector = new float[rsize];
        float * index = new float[rsize];
        
        for(int n=0; n< niter;n++){
            
            for(int x=0;x<width;x++)
                for(int y=0;y<height;y++){
                    
                    int count=0;
                    for(int i=-iradius;i<=iradius;i++)
                        for(int j=-iradius;j<=iradius;j++)
                            if ((float) (i*i + j*j) <= iradius*iradius){
                                
                                int x0=x+i;
                                int y0=y+j;
                                
                                if (x0>=0 && y0>=0 && x0 < width && y0 < height) { 
                                    vector[count] = u[y0*width+x0];
                                    index[count] = count;
                                    count++; 
                                }
                            }
                    
                    quick_sort(vector,index,count);
                    v[y*width+x] = vector[count/2];
                    
                }
            
            copy(v,u,width*height);
        }
        
        delete[] vector;
        delete[] index;
        
    }
    
    
    
    
    
    /* Convolution with a  separable kernel */
    /* boundary condition: 0=zero, 1=symmetry */
    void separable_convolution(float *u, float *v, int width, int height,float * xkernel, int xsize,float *ykernel,int ysize,int boundary)
    {
        
        int width2 = 2*width;
        int height2 = 2*height;
        
        float *tmp = (float *) malloc(width*height*sizeof(float));
        
        
        /* convolution along x axis */
        float sum = 0.0;
        int org = xsize / 2;
        for (int y=height;y--;) 
            for (int x=width;x--;) {
                
                sum = 0.0;	
                for (int i=xsize;i--;) {
                    int s = x-i+org;
                    switch(boundary) {
                            
                        case 0: 
                            if (s>=0 && s<width) sum += xkernel[i]*u[y*width+s];
                            break;
                            
                        case 1: 
                            
                            while (s<0) s+=width2;
                            while (s>=width2) s-=width2;
                            if (s>=width) s = width2-1-s;
                            sum += xkernel[i]*u[y*width+s];
                            break;
                            
                    }
                }
                tmp[y*width+x] = sum;
            }
        
        /* convolution along y axis */
        org = ysize / 2;
        for (int y=height;y--;) 
            for (int x=width;x--;) {
                
                sum=0.0;
                for (int i=ysize;i--;) {
                    int s = y-i+org;
                    switch(boundary) {
                        case 0: 
                            if (s>=0 && s<height) sum += ykernel[i]*tmp[s*width+x];
                            break;
                        case 1: 
                            while (s<0) s+=height2;
                            while (s>=height2) s-=height2;
                            if (s>=height) s = height2-1-s;
                            sum += ykernel[i]*tmp[s*width+x];
                            break;
                    }
                }
                v[y*width+x] = sum;
            }
        
        free(tmp);
    }
    
    
    void gaussian_convolution(float *u, float *v, int width, int height, float sigma)
    {
        
        int ksize;	
        float * kernel;
        
        ksize = (int)(2.0 * 4.0 * sigma + 1.0);
        kernel = gauss(1,sigma,&ksize);
        
        int boundary = 1;
        
        copy(u,v,width*height);
        horizontal_convolution(v, v, width, height, kernel, ksize, boundary);
        vertical_convolution(v, v, width, height,  kernel,  ksize, boundary);
        delete[] kernel; /*memcheck*/
    }
    
    
    void gaussian_convolution(float *u, float *v, int width, int height, float sigma, int ksize)
    {
        float * kernel;
        kernel = gauss(1,sigma,&ksize);
        
        int boundary = 1;
        
        copy(u,v,width*height);
        horizontal_convolution(v, v, width, height, kernel, ksize, boundary);
        vertical_convolution(v, v, width, height,  kernel,  ksize, boundary);
    }
    
    
    void fast_separable_convolution(float *u, float *v, int width, int height,float * xkernel, int xsize,float *ykernel,int ysize,int boundary)
    {
        copy(u,v,width*height);
        
        horizontal_convolution(v, v, width, height, xkernel, xsize, boundary);
        vertical_convolution(v, v, width, height,  ykernel,  ysize, boundary);
        
    }
    
    /* Loop unrolling simply sums 5 multiplications 
     at a time to allow the compiler to schedule
     operations better and avoid loop overhead.
     */
    void buffer_convolution(float *buffer,float *kernel,int size,int ksize)
    {
        
        for (int i = 0; i < size; i++) {
            
            float sum = 0.0;
            float *bp = &buffer[i];
            float *kp = &kernel[0];
            
            
            /* Loop unrolling: do 5 multiplications at a time. */
            //      int k=0;
            
            for(int k = 0; k < ksize; k++) 
                sum += *bp++ * *kp++;
            
            //    for(;k + 4 < ksize;  bp += 5, kp += 5, k += 5) 
            //	        sum += bp[0] * kp[0] +  bp[1] * kp[1] + bp[2] * kp[2] +
            //      	       bp[3] * kp[3] +  bp[4] * kp[4];
            
            /* Do multiplications at a time on remaining items. */
            //      for(; k < ksize; bp++ , kp++, k++)  sum += *bp * (*kp);
            
            buffer[i] = sum;
        }
    }
    
    
    
    /* Convolve image with the 1-D kernel vector along image rows.  This
     is designed to be as efficient as possible.  
     */
    void horizontal_convolution(float *u, float *v, int width, int height, float *kernel, int ksize, int boundary)
    {
        
        int halfsize = ksize / 2;
        int buffersize = width + ksize;
        float *buffer = new float[buffersize];
        
        for (int r = 0; r < height; r++) {
            
            /// symmetry
            int l = r*width;
            if (boundary == 1)
                for (int i = 0; i < halfsize; i++)
                    buffer[i] = u[l + halfsize - 1 - i ];
            else
                for (int i = 0; i < halfsize; i++)
                    buffer[i] = 0.0;
            
            
            for (int i = 0; i < width; i++)
                buffer[halfsize + i] = u[l + i];
            
            
            if (boundary == 1)
                for (int i = 0; i <  halfsize; i++)
                    buffer[i + width + halfsize] = u[l + width - 1 - i];
            else 
                for (int i = 0; i <  halfsize; i++)
                    buffer[i + width + halfsize] = 0.0;
            
            buffer_convolution(buffer, kernel, width, ksize);
            for (int c = 0; c < width; c++)
                v[r*width+c] = buffer[c];
        }
        delete[] buffer; /*memcheck*/
    }
    
    
    
    void vertical_convolution(float *u, float *v, int width, int height, float *kernel,int ksize, int boundary)
    {
        int halfsize = ksize / 2;
        int buffersize = height + ksize;
        float *buffer = new float[buffersize];
        
        for (int c = 0; c < width; c++) {
            
            if (boundary == 1)
                for (int i = 0; i < halfsize; i++)
                    buffer[i] = u[(halfsize-i-1)*width + c];
            else 
                for (int i = 0; i < halfsize; i++)
                    buffer[i] = 0.0f;
            
            for (int i = 0; i < height; i++)
                buffer[halfsize + i] = u[i*width + c];
            
            if (boundary == 1)
                for (int i = 0; i < halfsize; i++)
                    buffer[halfsize + height + i] = u[(height - i - 1)*width+c];
            else
                for (int i = 0; i < halfsize; i++)
                    buffer[halfsize + height + i] = 0.0f;
            
            buffer_convolution(buffer, kernel, height, ksize);
            
            for (int r = 0; r < height; r++)
                v[r*width+c] = buffer[r];
            
        }
        delete[] buffer; /*memcheck*/
    }
    
    
    
    void heat(float *input, float *out, float step, int niter, float sigma, int width, int height)
    {
        
        int i,j,n,ksize,size,im,i1,j1,jm;
        float *kernel = NULL, *laplacian = NULL, *convolved = NULL;
        
        
        size = width*height;
        
        if (sigma > 0.0) kernel = gauss(0,sigma,&ksize);		  		
        
        laplacian = (float *) malloc(size*sizeof(float));
        convolved = (float *) malloc(size*sizeof(float));
        
        
        
        for(n=0; n < niter; n++)
        {
            
            
            if (sigma > 0.0)
            {
                
                separable_convolution(input,convolved,width,height, kernel, ksize,kernel,ksize,1);
                
                for(i=0; i< size; i++) laplacian[i] = convolved[i] - input[i];
                
            } else
            {
                
                
                for (i=0; i < width;i++)
                    for (j=0; j< height ;j++) 
                    {
                        
                        if (j==0) jm=1; else jm=j-1;
                        if (j==height-1) j1=height-2; else j1=j+1;
                        
                        if (i==0) im=1; else im=i-1;
                        if (i==width-1) i1=width-2; else i1=i+1;
                        
                        laplacian[j*width + i] =  - 4.0  * input[width*j+i] + input[width*j+im]+ input[width*j+i1]+input[width*jm + i] + input[width*j1 + i];
                    }
            }
            
            
            
            for(i=0; i < size; i++) out[i] = input[i] + step * laplacian[i];
            
            copy(out,input,size);
            
        }
        
        
        free(laplacian);
        free(convolved);
        if (kernel) free(kernel);
        
    }
    
    
    
    
    
    
    
    float **allocate_float_matrix(int nrows, int ncols)
    {
        float ** matrix;
        matrix = new float*[nrows];
        for(int i=0; i < nrows; i++) matrix[i] = new float[ncols];
        return matrix;
    }
    
    
    void desallocate_float_matrix(float **matrix, int nrows, int ncols)
    {
        if (matrix == NULL) return;
        
        for(int i=0; i < nrows; i++) { delete[] matrix[i]; matrix[i] = 0;}
        matrix = 0;
        
        ncols = ncols; // to remove the warning unused parameter °Æncols°Ø
    }
    
    
    // **********************************************
    //  LU based algorithms
    // **********************************************
    
    
    
    // Solves Ax=b by using lu decomposition 
    int lusolve(float **a, float *x, float *b, int n)
    {
        
        float d;
        int *indx = new int[n];
        
        if (ludcmp(a,n,indx,&d))
        {
            
            for(int i=0; i < n; i++) x[i] = b[i];
            lubksb(a,n,indx,x);
            
            delete[] indx; /*memcheck*/
            return 1;
        } else
        {
            
            printf("lusolve::lu decomposition failed\n");
            delete[] indx; /*memcheck*/
            return 0;
        }	
        delete[] indx; // Guoshen Yu
    }
    
    
    
    int ludcmp(float **a, int n, int *indx, float *d)
    {
        
        int i,imax=0,j,k,aux;
        float big,dum,sum,temp;
        float *vv;
        
        
        vv=(float *) malloc(n*sizeof(float));
        *d=1.0;
        
        
        for(i=0;i<n;i++) indx[i]=i;
        
        
        /**** Look for the largest value of every line and store 1/this value****/
        for(i=0;i<n;i++){
            
            big=0.0;
            for(j=0;j<n;j++) 
                if ( (temp=fabs(a[i][j]))>big ) big=temp;
            
            if (big==0.0) { return 0; printf("LU Decomposition failed\n");}
            
            vv[i]=1.0/big;
        }
        
        
        
        
        for(j=0;j<n;j++){
            
            for(i=0;i<j;i++){
                
                sum=a[i][j];
                for(k=0;k<i;k++) sum-= a[i][k]*a[k][j];
                a[i][j]=sum; 
                
            }
            
            
            big=0.0;
            for(i=j;i<n;i++){
                
                sum=a[i][j];
                for(k=0;k<j;k++)   
                    sum -=a[i][k]*a[k][j];
                a[i][j]=sum;
                
                if ( (dum=vv[i]*fabs(sum))>=big){
                    big=dum;
                    imax=i;
                }
            }
            
            
            if (j != imax){
                
                for(k=0;k<n;k++){
                    
                    dum=a[imax][k];
                    a[imax][k]=a[j][k];
                    a[j][k]=dum;
                }
                
                *d=-(*d);
                vv[imax]=vv[j];
                
                aux=indx[j];
                indx[j]=indx[imax];
                indx[imax]=aux;
            }
            
            if (a[j][j]==0.0)  a[j][j]=NRTINY;
            
            
            if (j!=n-1 ){
                
                dum=1.0 / a[j][j];
                for(i=j+1;i<n;i++) a[i][j]*=dum;
            }
            
            
        }
        
        free(vv);
        return 1;
    }
    
    
    
    /* Solves the set of n linear equations Ax=b. Here a[0..n-1][0..n-1] as input, not as the matrix A but rather as its LU decomposition,*/
    /* determined by the routine ludcmp. indx[0..n-1] is input as the permutation vector returned by ludcmp. b[0..n-1] is input as the    */
    /* right hand side vector and returns with the solution vector x. */
    void lubksb(float **a, int n, int *indx, float *b)
    {
        
        int i,ii=0,j;
        float sum,*aux;
        
        aux= (float *) malloc(n*sizeof(float));
        
        
        for(i=0;i<n;i++)
            aux[i]=b[indx[i]];
        
        
        for(i=0;i<n;i++){
            sum=aux[i];
            
            if (ii)
                for(j=ii-1;j<i;j++) sum-=a[i][j]*aux[j];
            else if (sum) ii=i+1;
            aux[i]=sum;
        }
        
        for(i=n-1;i>=0;i--){
            sum=aux[i];
            for(j=i+1;j<n;j++) sum-=a[i][j]*aux[j];
            aux[i]=sum/a[i][i];
            b[i]=sum/a[i][i];
        }
        
        free(aux);
    }
    
    
    
    
    
    
    
    void default_sift_parameters(siftPar &par)
    {
        par.OctaveMax=100000;
        par.DoubleImSize = 0;
        par.order = 3;
        par.InitSigma = 1.6;
        par.BorderDist = 5; 
        par.Scales = 3;
        par.PeakThresh = 255.0 * 0.04 / 3.0;
        par.EdgeThresh = 0.06; 
        par.EdgeThresh1 = 0.08; 
        par.OriBins  = 36;
        par.OriSigma = 1.5;
        par.OriHistThresh = 0.8;
        par.MaxIndexVal = 0.2;
        par.MagFactor  = 3;
        par.IndexSigma  = 1.0;
        par.IgnoreGradSign = 0;
        //	par.MatchRatio = 0.6;
        //	par.MatchRatio = 0.75; // Guoshen Yu. Since l1 distance is used for matching instead of l2, a larger threshold is needed. 
        par.MatchRatio = 0.73; // Guoshen Yu. Since l1 distance is used for matching instead of l2, a larger threshold is needed. 
        par.MatchXradius = 1000000.0f;
        par.MatchYradius = 1000000.0f;
        
        par.noncorrectlylocalized = 0;
        
    }
    
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /// SIFT Keypoint detection 
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    
    
    
    void OctaveKeypoints(flimage & image, float octSize, keypointslist& keys,siftPar &par);
    
    void FindMaxMin(  flimage* dogs,  flimage* blur, float octSize ,keypointslist& keys,siftPar &par);
    
    bool LocalMax(float val, flimage& dog, int y0, int x0);
    
    bool LocalMin(float val, flimage& dog, int y0, int x0);
    
    bool LocalMaxMin(float val, const flimage& dog, int y0, int x0);
    
    int NotOnEdge( flimage& dog, int r, int c, float octSize,siftPar &par);
    
    float FitQuadratic(float offset[3],  flimage* dogs, int s, int r, int c);
    
    void InterpKeyPoint(
                        flimage* dogs, int s, int r, int c,
                        const flimage& grad, const flimage& ori, flimage& map,
                        float octSize, keypointslist& keys, int movesRemain,siftPar &par);
    
    void AssignOriHist(
                       const flimage& grad, const flimage& ori, float octSize,
                       float octScale, float octRow, float octCol, keypointslist& keys,siftPar &par);
    
    void SmoothHistogram(
                         float* hist, int bins);
	
    float InterpPeak(
                     float a, float b, float c);
    
    void MakeKeypoint(
                      const flimage& grad, const flimage& ori, float octSize, float octScale,
                      float octRow, float octCol, float angle, keypointslist& keys,siftPar &par);
    
    void MakeKeypointSample(
                            keypoint& key, const flimage& grad, const flimage& ori,
                            float scale, float row, float col,siftPar &par);
	
    void NormalizeVec(
                      float* vec);
	
    void KeySampleVec(
                      keypoint& key, const flimage& grad, const flimage& ori,
                      float scale, float row, float col,siftPar &par);
	
    void KeySample(
                   float index[IndexSize][IndexSize][OriSize], keypoint& key,
                   const flimage& grad, const flimage& ori,
                   float scale, float row, float col,siftPar &par);
    
    void AddSample(
                   float index[IndexSize][IndexSize][OriSize], keypoint& key,
                   const flimage& grad, const flimage& orim,
                   float r, float c, float rpos, float cpos, float rx, float cx,siftPar &par);
    
    void PlaceInIndex(
                      float index[IndexSize][IndexSize][OriSize],
                      float mag, float ori, float rx, float cx,siftPar &par);
    
    
    
    
    void compute_sift_keypoints(float *input, keypointslist& keypoints, int width, int height, siftPar &par)
    {
        
        flimage image;
        
        /// Make zoom of image if necessary
        float octSize = 1.0;
        if (par.DoubleImSize){
            
            //printf("... compute_sift_keypoints :: applying zoom\n");
            //		image.create(2*width, 2*height);
            //		apply_zoom(input,image.getPlane(),2.0,par.order,width,height);
            //		octSize *= 0.5;
            
            printf("Doulbe image size not allowed. Guoshen Yu\n");
            exit(-1);
            
            
            
        } else
        {
            
            image.create(width,height,input);
        }
        
        // 	 printf("Using initial Dog value: %f\n", par.PeakThresh);
        //    	 printf("Double image size: %d\n", par.DoubleImSize);
        //    	 printf("Interpolation order: %d\n", par.order);
        
        
        /// Apply initial smoothing to input image to raise its smoothing to par.InitSigma.  
        /// We assume image from camera has smoothing of sigma = 0.5, which becomes sigma = 1.0 if image has been doubled. 
        /// increase = sqrt(Init^2 - Current^2)
        float	curSigma;
        if (par.DoubleImSize) curSigma = 1.0; else curSigma = 0.5;
        
        
        if (par.InitSigma > curSigma ) {
            
            if (DEBUG) printf("Convolving initial image to achieve std: %f \n", par.InitSigma);
            
            float sigma = (float) sqrt((double)(par.InitSigma * par.InitSigma - curSigma * curSigma));
            
            gaussian_convolution( image.getPlane(), image.getPlane(), image.nwidth(), image.nheight(), sigma);
            
        }
        
        
        
        /// Convolve by par.InitSigma at each step inside OctaveKeypoints by steps of 
        /// Subsample of factor 2 while reasonable image size 
        
        /// Keep reducing image by factors of 2 until one dimension is
        /// smaller than minimum size at which a feature could be detected.
        int 	minsize = 2 * par.BorderDist + 2;
        int     OctaveCounter = 0;
        //printf("... compute_sift_keypoints :: maximum number of scales : %d\n", par.OctaveMax);
        
        while (image.nwidth() > minsize &&  image.nheight() > minsize && OctaveCounter < par.OctaveMax) {
            
            if (DEBUG) printf("Calling OctaveKeypoints \n");
            
            OctaveKeypoints(image, octSize, keypoints,par); 
            
            // image is blurred inside OctaveKeypoints and therefore can be sampled
            flimage aux( (int)((float) image.nwidth() / 2.0f) , (int)((float) image.nheight() / 2.0f)); 
            
            if (DEBUG) printf("Sampling initial image \n");
            
            sample(image.getPlane(), aux.getPlane(), 2.0f, image.nwidth(), image.nheight());
            
            image = aux;
            
            octSize *= 2.0;
            
            OctaveCounter++;
            
        }
        
        
        /*	printf("sift::  %d keypoints\n", keypoints.size());
         printf("sift::  plus non correctly localized: %d \n", 	par.noncorrectlylocalized);*/
        
    }
    
    
    
    /////////////////////////////////////////////////
    /// EXTREMA DETECTION IN ONE SCALE-SPACE OCTAVE:
    /////////////////////////////////////////////////
    
    /// par.Scales determine how many steps we perform to pass from one scale to the next one:  sigI --> 2*sigI
    /// At each step we pass from    sigma_0 --> sigma0 * (1 + R)
    /// At the last step  sigI * (1 + R)^par.Scales = 2 * sigI
    /// (1+R) = 2 ^(1 / par.Scales)  it is called sigmaRatio
    
    /// It seems that blur[par.Scales+1] is compared in two succesive iterations
    void OctaveKeypoints(flimage & image, float octSize, keypointslist& keys,siftPar &par)
    {
        // Guoshen Yu, 2010.09.21, Windows version
        // flimage blur[par.Scales+3], dogs[par.Scales+2];
        int size_blur = par.Scales+3;
        int size_dogs = par.Scales+2;	  
        flimage *blur = new flimage[size_blur];
        flimage *dogs = new flimage[size_dogs];
        
        float sigmaRatio = (float) pow(2.0, 1.0 / (double) par.Scales);
        
        
        /* Build array, blur, holding par.Scales+3 blurred versions of the image. */
        blur[0] = flimage(image);	/* First level is input to this routine. */ 
        float prevSigma = par.InitSigma;	/* Input image has par.InitSigma smoothing. */
        
        
        /* Form each level by adding incremental blur from previous level.
         Increase in blur is from prevSigma to prevSigma * sigmaRatio, so
         increase^2 = (prevSigma * sigmaRatio)^2 - prevSigma^2 
         */
        for (int i = 1; i < par.Scales + 3; i++) {
            
            if (DEBUG) printf("Convolving scale: %d \n", i);
            
            blur[i] = flimage(blur[i-1]);
            
            float increase = prevSigma*(float)sqrt((double)(sigmaRatio*sigmaRatio-1.0));
            
            gaussian_convolution( blur[i].getPlane(), blur[i].getPlane(), blur[i].nwidth(), blur[i].nheight(), increase);
            
            prevSigma *= sigmaRatio;
            
        }
        
        
        /* Compute an array, dogs, of difference-of-Gaussian images by
         subtracting each image from its next blurred version. */
        for (int i = 0; i < par.Scales + 2; i++) {
            
            dogs[i] = flimage(blur[i]);
            
            /// dogs[i] = dogs[i] - blur[i+1]
            combine(dogs[i].getPlane(),1.0f, blur[i+1].getPlane(),-1.0f, dogs[i].getPlane(),  dogs[i].nwidth() * dogs[i].nheight()); 
        }
        
        
        // Image with exact blur to be subsampled is blur[scales]
        image = blur[par.Scales];
        
        /* Scale-space extrema detection in this octave	*/
        if (DEBUG) printf("Looking for local maxima \n");
        
        FindMaxMin(dogs, blur, octSize, keys,par);
        
        // Guoshen Yu, 2010.09.22, Windows version
        delete [] blur;
        delete [] dogs;
    }
	
    
    /////////////////////////////////////////////////
    ///Find the local maxima and minima of the DOG images in scale space.  Return the keypoints for these locations.
    /////////////////////////////////////////////////
    
    /// For each point at each scale we decide if it is a local maxima:
    /// - |dogs(x,y,s)| > 0.8 * par.PeakThresh
    /// - Local maximum or minimum in s-1,s,s+1
    /// - NotonEdge:  ratio of the two principle curvatures of the DOG function at this point be below a threshold.
    
    
    /// blur[par.Scales+1] is not used in order to look for extrema
    /// while these could be computed using avalaible blur and dogs
    void FindMaxMin(
                    flimage* dogs,  flimage* blur,
                    float octSize, keypointslist& keys,siftPar &par)
    {
        
        int width = dogs[0].nwidth(), height = dogs[0].nheight();
        
        /* Create an image map in which locations that have a keypoint are
         marked with value 1.0, to prevent two keypoints being located at
         same position.  This may seem an inefficient data structure, but
         does not add significant overhead.
         */
        flimage map(width,height,0.0f);
        flimage grad(width,height,0.0f);
        flimage ori(width,height,0.0f);
        
        /* Search through each scale, leaving 1 scale below and 1 above.
         There are par.Scales+2 dog images.
         */
        for (int s = 1; s < par.Scales+1; s++) {
            
            if (DEBUG) printf("************************scale: %d\n", s);
            
            //getchar();
            
            /* For each intermediate image, compute gradient and orientation
             images to be used for keypoint description.  */
            compute_gradient_orientation(blur[s].getPlane(), grad.getPlane(), ori.getPlane(), blur[s].nwidth(), blur[s].nheight());
            
            
            /* Only find peaks at least par.BorderDist samples from image border, as
             peaks centered close to the border will lack stability. */
            assert(par.BorderDist >= 2);
            float val;
            int partialcounter = 0;
            for (int r = par.BorderDist; r < height - par.BorderDist; r++) 
                for (int c = par.BorderDist; c < width - par.BorderDist; c++) {
                    
                    /* Pixel value at (c,r) position. */
                    val = dogs[s](c,r);	
                    
                    /* DOG magnitude must be above 0.8 * par.PeakThresh threshold
                     (precise threshold check will be done once peak
                     interpolation is performed).  Then check whether this
                     point is a peak in 3x3 region at each level, and is not
                     on an elongated edge.
                     */
                    
                    if (fabs(val) > 0.8 * par.PeakThresh)
                    {
                        
                        /*
                         
                         // If local maxima
                         if (LocalMax(val, dogs[s-1], r, c,par) && LocalMax(val, dogs[s], r, c, par) && LocalMax(val, dogs[s+1], r, c,par) && NotOnEdge(dogs[s], r, c, octSize,par))
                         {
                         if (DEBUG) printf("Maximum Keypoint found (%d,%d,%d)  val: %f\n",s,r,c,val);
                         InterpKeyPoint(
                         dogs, s, r, c, grad, ori,
                         map, octSize, keys, 5,par);	
                         
                         } else  if (LocalMin(val, dogs[s-1], r, c,par) && LocalMin(val, dogs[s], r, c,par) && LocalMin(val, dogs[s+1], r, c,par) && NotOnEdge(dogs[s], r, c, octSize,par))
                         {
                         if (DEBUG) printf("Minimum Keypoint found (%d,%d,%d)  val: %f\n",s,r,c,val);
                         InterpKeyPoint(
                         dogs, s, r, c, grad, ori,
                         map, octSize, keys, 5,par);	
                         }
                         */
                        if (LocalMaxMin(val, dogs[s-1], r, c) && LocalMaxMin(val, dogs[s], r, c) && LocalMaxMin(val, dogs[s+1], r, c) && NotOnEdge(dogs[s], r, c, octSize,par))
                        {
                            partialcounter++;
                            if (DEBUG) printf("%d:  (%d,%d,%d)  val: %f\n",partialcounter, s,r,c,val);
                            
                            InterpKeyPoint(
                                           dogs, s, r, c, grad, ori,
                                           map, octSize, keys, 5,par);	
                            
                            //getchar();
                        } 
                        
                        
                    }
                    
                }
        }
        
    }
    
    
    //bool LocalMax(float val, flimage& dog, int y0, int x0, siftPar &par)
    bool LocalMax(float val, flimage& dog, int y0, int x0)
    {
        for (int x = x0 - 1; x <= x0 + 1; x++)
            for (int y = y0  - 1; y <= y0 + 1; y++){
                //printf("%f \t", dog(x,y));
                if (dog(x,y) > val) return 0;
            }
        
        return 1;
    }
    
    bool LocalMin(float val,  flimage& dog, int y0, int x0)
    {
        for (int x = x0 - 1; x <= x0 + 1; x++)
            for (int y = y0  - 1; y <= y0 + 1; y++){
                //printf("%f \t", dog(x,y));
                if (dog(x,y) < val) return 0;
            }
        
        return 1;
    }
    
    
    /* Return TRUE iff val is a local maximum (positive value) or
     minimum (negative value) compared to the 3x3 neighbourhood that
     is centered at (row,col).
     */
    bool LocalMaxMin(float val, const flimage& dog, int y0, int x0)
    {
        // For efficiency, use separate cases for maxima or minima, and
        // return as soon as possible
        if (val > 0.0) {
            for (int x = x0 - 1; x <= x0 + 1; x++)
                for (int y = y0  - 1; y <= y0 + 1; y++){
                    if (dog(x,y) > val) return false;
                }
        } else {
            for (int x = x0 - 1; x <= x0 + 1; x++)
                for (int y = y0  - 1; y <= y0 + 1; y++){
                    if (dog(x,y) < val) return false;
                }
        }
        
        return true;
    }
    
    
    
    /* Returns FALSE if this point on the DOG function lies on an edge.
     This test is done early because it is very efficient and eliminates
     many points.  It requires that the ratio of the two principle
     curvatures of the DOG function at this point be below a threshold.
     
     Edge threshold is higher on the first scale where SNR is small in 
     order to reduce the number of unstable keypoints. 
     */
    int NotOnEdge(flimage& dog, int r, int c, float octSize,siftPar &par)
    {
        /* Compute 2x2 Hessian values from pixel differences. */
        float	H00 = dog(c,r-1) - 2.0 * dog(c,r) + dog(c,r+1), /* AMIR: div by ? */
		H11 = dog(c-1,r) - 2.0 * dog(c,r) + dog(c+1,r),
		H01 = ( (dog(c+1,r+1) - dog(c-1,r+1)) - (dog(c+1,r-1) - dog(c-1,r-1)) ) / 4.0;
        
        /* Compute determinant and trace of the Hessian. */
        float	det = H00 * H11 - H01 * H01,	/// Det H = \prod l_i
		trace = H00 + H11;		/// tr H = \sum l_i	
        
        /// As we do not desire edges but only corners we demand l_max / l_min less than a threshold
        /// In practice if A = k B,     A*B = k B^2
        ///				(A + B)^2 = (k+1)^2 * B^2
        ///				k B^2 >  t * (k+1)^2 * B^2 sii   k  / (k+1)^2 > t
        /// This is a decreasing function for k > 1 and value 0.3 at k=1.
        /// Setting t = 0.08, means k<=10 
        
        /* To detect an edge response, we require the ratio of smallest
         to largest principle curvatures of the DOG function
         (eigenvalues of the Hessian) to be below a threshold.  For
         efficiency, we use Harris' idea of requiring the determinant to
         be above par.EdgeThresh times the squared trace, as for eigenvalues
         A and B, det = AB, trace = A+B.  So if A = 10B, then det = 10B**2,
         and trace**2 = (11B)**2 = 121B**2, so par.EdgeThresh = 10/121 =
         0.08 to require ratio of eigenvalues less than 10.
         */
        if (octSize <= 1)
            return (det > par.EdgeThresh1 * trace * trace);
        else
            return (det > par.EdgeThresh * trace * trace);
        
    }
    
    
    /* Create a keypoint at a peak near scale space location (s,r,c), where
     s is scale (index of DOGs image), and (r,c) is (row, col) location.
     Add to the list of keys with any new keys added.
     */
    void InterpKeyPoint(
                        flimage* dogs, int s, int r, int c,
                        const flimage& grad, const flimage& ori, flimage& map,
                        float octSize, keypointslist& keys, int movesRemain,siftPar &par)
    {
        
        /* Fit quadratic to determine offset and peak value. */
        float offset[3];
        float peakval = FitQuadratic(offset, dogs, s, r, c);
        if (DEBUG) printf("peakval: %f, of[0]: %f  of[1]: %f  of[2]: %f\n", peakval, offset[0], offset[1], offset[2]);
        
        /* Move to an adjacent (row,col) location if quadratic interpolation
         is larger than 0.6 units in some direction (we use 0.6 instead of
         0.5 to avoid jumping back and forth near boundary).  We do not
         perform move to adjacent scales, as it is seldom useful and we
         do not have easy access to adjacent scale structures.  The
         movesRemain counter allows only a fixed number of moves to
         prevent possibility of infinite loops.
         */
        int newr = r, newc = c;
        if (offset[1] > 0.6 && r < dogs[0].nheight() - 3)
            newr++;
        else if (offset[1] < -0.6 && r > 3)
            newr--;
        
        if (offset[2] > 0.6 && c < dogs[0].nwidth() - 3)
            newc++;
        else if (offset[2] < -0.6 && c > 3)
            newc--;
        
        if (movesRemain > 0  &&  (newr != r || newc != c)) {
            InterpKeyPoint(
                           dogs, s, newr, newc, grad, ori, map,
                           octSize, keys,movesRemain - 1,par);
            return;
        }
        
        /* Do not create a keypoint if interpolation still remains far
         outside expected limits, or if magnitude of peak value is below
         threshold (i.e., contrast is too low). */
        if (	fabs(offset[0]) > 1.5 || fabs(offset[1]) > 1.5 ||
            fabs(offset[2]) > 1.5 || fabs(peakval) < par.PeakThresh)		
		{
			if (DEBUG) printf("Point not well localized by FitQuadratic\n"); 	
			par.noncorrectlylocalized++;
			return;
		}
        
        /* Check that no keypoint has been created at this location (to avoid
         duplicates).  Otherwise, mark this map location.
         */
        if (map(c,r) > 0.0) return;
        map(c,r) = 1.0;
        
        /* The scale relative to this octave is given by octScale.  The scale
         units are in terms of sigma for the smallest of the Gaussians in the
         DOG used to identify that scale.
         */
        // Guoshen Yu, 2010.09.21 Windows version
        // float octScale = par.InitSigma * pow(2.0, (s + offset[0]) / (float) par.Scales);
        float octScale = par.InitSigma * pow(2.0, (s + offset[0]) / (double) par.Scales);
        
        /// always use histogram of orientations
        //if (UseHistogramOri)
		AssignOriHist(
                      grad, ori, octSize, octScale,
                      r + offset[1], c + offset[2], keys, par);
        //else
        //	AssignOriAvg(
        //		grad, ori, octSize, octScale,
        //		r + offset[1], c + offset[2], keys);
    }
    
    
    
    /* Apply the method developed by Matthew Brown (see BMVC 02 paper) to
     fit a 3D quadratic function through the DOG function values around
     the location (s,r,c), i.e., (scale,row,col), at which a peak has
     been detected.  Return the interpolated peak position as a vector
     in "offset", which gives offset from position (s,r,c).  The
     returned value is the interpolated DOG magnitude at this peak.
     */
    float FitQuadratic(float offset[3], flimage* dogs, int s, int r, int c)
    {
        float g[3];
        flimage *dog0, *dog1, *dog2;
        int i;
        
        //s = 1; r = 128; c = 128;
        
        float ** H =  allocate_float_matrix(3, 3);
        
        /* Select the dog images at peak scale, dog1, as well as the scale
         below, dog0, and scale above, dog2. */
        dog0 = &dogs[s-1];
        dog1 = &dogs[s];
        dog2 = &dogs[s+1];
        
        /* Fill in the values of the gradient from pixel differences. */
        g[0] = ((*dog2)(c,r) - (*dog0)(c,r)) / 2.0;
        g[1] = ((*dog1)(c,r+1) - (*dog1)(c,r-1)) / 2.0;
        g[2] = ((*dog1)(c+1,r) - (*dog1)(c-1,r)) / 2.0;
        
        /* Fill in the values of the Hessian from pixel differences. */
        H[0][0] = (*dog0)(c,r)   - 2.0 * (*dog1)(c,r) + (*dog2)(c,r);
        H[1][1] = (*dog1)(c,r-1) - 2.0 * (*dog1)(c,r) + (*dog1)(c,r+1);
        H[2][2] = (*dog1)(c-1,r) - 2.0 * (*dog1)(c,r) + (*dog1)(c+1,r);
        H[0][1] = H[1][0] = ( ((*dog2)(c,r+1) - (*dog2)(c,r-1)) -
                             ((*dog0)(c,r+1) - (*dog0)(c,r-1)) ) / 4.0;
        H[0][2] = H[2][0] = ( ((*dog2)(c+1,r) - (*dog2)(c-1,r)) -
                             ((*dog0)(c+1,r) - (*dog0)(c-1,r)) ) / 4.0;
        H[1][2] = H[2][1] = ( ((*dog1)(c+1,r+1) - (*dog1)(c-1,r+1)) -
                             ((*dog1)(c+1,r-1) - (*dog1)(c-1,r-1)) ) / 4.0;
        
        /* Solve the 3x3 linear sytem, Hx = -g. Result, x, gives peak offset.
         Note that SolveLinearSystem destroys contents of H. */
        offset[0] = - g[0];
        offset[1] = - g[1];
        offset[2] = - g[2];
        
        // 	for(i=0; i < 3; i++){
        // 	
        // 		for(j=0; j < 3; j++) printf("%f  ", H[i][j]);
        // 		printf("\n");
        // 	}
        
        // 	printf("\n");
        // 
        // 	for(i=0; i < 3; i++) printf("%f  ", offset[i]);
        // 	printf("\n");
        
        float solution[3];
        lusolve(H, solution, offset,3);
        
        // 	printf("\n");
        // 	for(i=0; i < 3; i++) printf("%f  ", solution[i]);
        // 	printf("\n");
        
        
        desallocate_float_matrix(H,3,3);
        delete[] H; /*memcheck*/
        
        /* Also return value of DOG at peak location using initial value plus
         0.5 times linear interpolation with gradient to peak position
         (this is correct for a quadratic approximation).
         */
        for(i=0; i < 3; i++) offset[i] = solution[i];
        
        return ((*dog1)(c,r) + 0.5 * (solution[0]*g[0]+solution[1]*g[1]+solution[2]*g[2]));
    }
    
    
    
    /// - Compute histogram of orientation in a neighborhood weighted by gradient and distance to center
    /// - Look for local (3-neighborhood) maximum with valuer larger or equal than par.OriHistThresh * maxval
    
    
    /* Assign an orientation to this keypoint.  This is done by creating a
     Gaussian weighted histogram of the gradient directions in the
     region.  The histogram is smoothed and the largest peak selected.
     The results are in the range of -PI to PI.
     */
    void AssignOriHist(
                       const flimage& grad, const flimage& ori, float octSize,
                       float octScale, float octRow, float octCol,keypointslist& keys,siftPar &par)
    {
        int	bin, prev, next;
        
        // Guoshen Yu, 2010.09.21 Windows version
        //	float	hist[par.OriBins], distsq, dif, gval, weight, angle, interp;
        float distsq, dif, gval, weight, angle, interp;
        int tmp_size = par.OriBins;
        float *hist = new float[tmp_size];
        
        float radius2, sigma2;		
        
        int	row = (int) (octRow+0.5),
		col = (int) (octCol+0.5),
		rows = grad.nheight(),
		cols = grad.nwidth();
        
        for (int i = 0; i < par.OriBins; i++) hist[i] = 0.0;
        
        /* Look at pixels within 3 sigma around the point and sum their
         Gaussian weighted gradient magnitudes into the histogram. */
        float	sigma = par.OriSigma * octScale;
        int	radius = (int) (sigma * 3.0);
        int rmin = MAX(0,row-radius);
        int cmin = MAX(0,col-radius);
        int rmax = MIN(row+radius,rows-2);
        int cmax = MIN(col+radius,cols-2);
        radius2 = (float)(radius * radius);
        sigma2 = 2.0*sigma*sigma;
        
        for (int r = rmin; r <= rmax; r++) {
            for (int c = cmin; c <= cmax; c++) {
                
				gval = grad(c,r);
				
				dif = (r - octRow);	distsq = dif*dif;
				dif = (c - octCol);	distsq += dif*dif;
                
				if (gval > 0.0  &&  distsq < radius2 + 0.5) {
                    
					weight = exp(- distsq / sigma2);
					
					/* Ori is in range of -PI to PI. */
					angle = ori(c,r);
					bin = (int) (par.OriBins * (angle + PI + 0.001) / (2.0 * PI));
					assert(bin >= 0 && bin <= par.OriBins);
					bin = MIN(bin, par.OriBins - 1);
					hist[bin] += weight * gval;
                    
				}
                
            }
        }
        
        
        /* Apply smoothing 6 times for accurate Gaussian approximation. */
        for (int i = 0; i < 6; i++)
            SmoothHistogram(hist, par.OriBins);
        
        /* Find maximum value in histogram. */
        float maxval = 0.0;
        for (int i = 0; i < par.OriBins; i++) 
            if (hist[i] > maxval) maxval = hist[i];
        
        /* Look for each local peak in histogram.  If value is within
         par.OriHistThresh of maximum value, then generate a keypoint. */
        for (int i = 0; i < par.OriBins; i++) {
            prev = (i == 0 ? par.OriBins - 1 : i - 1);
            next = (i == par.OriBins - 1 ? 0 : i + 1);
            
            if (	hist[i] > hist[prev]  &&  hist[i] > hist[next]  &&
                hist[i] >= par.OriHistThresh * maxval ) {
                
                /* Use parabolic fit to interpolate peak location from 3 samples.
                 Set angle in range -PI to PI. */
                interp = InterpPeak(hist[prev], hist[i], hist[next]);
                angle = 2.0 * PI * (i + 0.5 + interp) / par.OriBins - PI;
                assert(angle >= -PI  &&  angle <= PI);
                
                if (DEBUG) printf("angle selected: %f \t location: (%f,%f)\n", angle, octRow, octCol);
                ;
                /* Create a keypoint with this orientation. */
                MakeKeypoint(
                             grad, ori, octSize, octScale,
                             octRow, octCol, angle, keys,par);
            }
            
        }
        
        // Guoshen Yu, 2010.09.22, Windows version
        delete [] hist;
    }
    
    
    
    /* Smooth a histogram by using a [1/3 1/3 1/3] kernel.  Assume the histogram
     is connected in a circular buffer.
     */
    void SmoothHistogram(float* hist, int bins)
    {
        float prev, temp;
        
        prev = hist[bins - 1];
        for (int i = 0; i < bins; i++) {
            temp = hist[i];
            hist[i] = ( prev + hist[i] + hist[(i + 1 == bins) ? 0 : i + 1] ) / 3.0;
            prev = temp;
        }
    }
    
    
    /* Return a number in the range [-0.5, 0.5] that represents the
     location of the peak of a parabola passing through the 3 evenly
     spaced samples.  The center value is assumed to be greater than or
     equal to the other values if positive, or less than if negative.
     */
    float InterpPeak(float a, float b, float c)
    {
        if (b < 0.0) {
            a = -a; b = -b; c = -c;
        }
        assert(b >= a  &&  b >= c);
        return 0.5 * (a - c) / (a - 2.0 * b + c);
    }
    
    
    
    
    
    /* Joan Pau: Add a new keypoint to a vector of keypoints
     Create a new keypoint and return list of keypoints with new one added.
     */
    void MakeKeypoint(
                      const flimage& grad, const flimage& ori, float octSize, float octScale,
                      float octRow, float octCol, float angle, keypointslist& keys,siftPar &par)
    {
        keypoint newkeypoint;
        newkeypoint.x = octSize * octCol;	/*x coordinate */
        newkeypoint.y = octSize * octRow;	/*y coordinate */
        newkeypoint.scale = octSize * octScale;	/* scale */
        newkeypoint.angle = angle;		/* orientation */
        MakeKeypointSample(newkeypoint,grad,ori,octScale,octRow,octCol,par);
        keys.push_back(newkeypoint);
    }
    
    
    
    /* Use the parameters of this keypoint to sample the gradient images
     at a set of locations within a circular region around the keypoint.
     The (scale,row,col) values are relative to current octave sampling.
     The resulting vector is stored in the key.
     */
    void MakeKeypointSample(
                            keypoint& key, const flimage& grad, const flimage& ori,
                            float scale, float row, float col,siftPar &par)
    {
        /* Produce sample vector. */
        KeySampleVec(key, grad, ori, scale, row, col,par);
        
        
        /* Normalize vector.  This should provide illumination invariance
         for planar lambertian surfaces (except for saturation effects).
         Normalization also improves nearest-neighbor metric by
         increasing relative distance for vectors with few features.
         It is also useful to implement a distance threshold and to
         allow conversion to integer format.
         */
        NormalizeVec(key.vec);
        
        /* Now that normalization has been done, threshold elements of
         index vector to decrease emphasis on large gradient magnitudes.
         Admittedly, the large magnitude values will have affected the
         normalization, and therefore the threshold, so this is of
         limited value.
         */
        bool changed = false;
        for (int i = 0; i < VecLength; i++)
            if (key.vec[i] > par.MaxIndexVal) {
                key.vec[i] = par.MaxIndexVal;
                changed = true;
            }
        
        if (changed) NormalizeVec(key.vec);
        
        /* Convert float vector to integer. Assume largest value in normalized
         vector is likely to be less than 0.5. */
        /// QUESTION: why is the vector quantized to integer
        int intval;
        for (int i = 0; i < VecLength; i++) {
            intval =  (int)(512.0 * key.vec[i]);
            key.vec[i] = (int) MIN(255, intval);
        }
    }
    
    /* Normalize length of vec to 1.0.
     */
    void NormalizeVec(float* vec)
    {
        float val, fac;
        
        float sqlen = 0.0;
        for (int i = 0; i < VecLength; i++) {
            val = vec[i];
            sqlen += val * val;
        }
        fac = 1.0 / sqrt(sqlen);
        
        for (int i = 0; i < VecLength; i++)
            vec[i] *= fac;
    }
    
    
    /* Create a 3D index array into which gradient values are accumulated.
     After filling array, copy values back into vec.
     */
    void KeySampleVec(
                      keypoint& key, const flimage& grad, const flimage& ori,
                      float scale, float row, float col,siftPar &par)
    {
        
        float index[IndexSize][IndexSize][OriSize];
        
        /* Initialize index array. */
        for (int i = 0; i < IndexSize; i++)
            for (int j = 0; j < IndexSize; j++)
                for (int k = 0; k < OriSize; k++)
                    index[i][j][k] = 0.0;
        
        
        KeySample(index, key, grad, ori, scale, row, col, par);
        
        
        /* Unwrap the 3D index values into 1D vec. */
        int v = 0;
        for (int i = 0; i < IndexSize; i++)
            for (int j = 0; j < IndexSize; j++)
                for (int k = 0; k < OriSize; k++)
                    key.vec[v++] = index[i][j][k];
    }
    
    
    
    /* Add features to vec obtained from sampling the grad and ori images
     for a particular scale.  Location of key is (scale,row,col) with respect
     to images at this scale.  We examine each pixel within a circular
     region containing the keypoint, and distribute the gradient for that
     pixel into the appropriate bins of the index array.
     */
    void KeySample(
                   float index[IndexSize][IndexSize][OriSize], keypoint& key,
                   const flimage& grad, const flimage& ori, float scale, float row, float col,siftPar &par)
    {
        float rpos, cpos, rx, cx;
        
        int	irow = (int) (row + 0.5),
		icol = (int) (col + 0.5);
        float	sine   = (float) sin(key.angle),
		cosine = (float) cos(key.angle);
        
        /* The spacing of index samples in terms of pixels at this scale. */
        float	spacing = scale * par.MagFactor;
        
        /* Radius of index sample region must extend to diagonal corner of
         index patch plus half sample for interpolation. */
        float	radius = 1.414 * spacing * (IndexSize + 1) / 2.0;
        int	iradius = (int) (radius + 0.5);
        
        /* Examine all points from the gradient image that could lie within the
         index square. */
        for (int i = -iradius; i <= iradius; i++) {
            for (int j = -iradius; j <= iradius; j++) {
                
                /* Rotate sample offset to make it relative to key orientation.
                 Uses (row,col) instead of (x,y) coords.  Also, make subpixel
                 correction as later image offset must be an integer.  Divide
                 by spacing to put in index units.
                 */
                
                /* Guoshen Yu, inverse the rotation */
                rpos = ((cosine * i - sine * j) - (row - irow)) / spacing;
                cpos = ((sine * i + cosine * j) - (col - icol)) / spacing;
                
                /*
                 rpos = ((cosine * i + sine * j) - (row - irow)) / spacing;
                 cpos = ((- sine * i + cosine * j) - (col - icol)) / spacing;*/
                
                /* Compute location of sample in terms of real-valued index array
                 coordinates.  Subtract 0.5 so that rx of 1.0 means to put full
                 weight on index[1] (e.g., when rpos is 0 and IndexSize is 3. */
                rx = rpos + IndexSize / 2.0 - 0.5;
                cx = cpos + IndexSize / 2.0 - 0.5;
                
                /* Test whether this sample falls within boundary of index patch. */
                if (	rx > -1.0 && rx < (float) IndexSize  &&
                    cx > -1.0 && cx < (float) IndexSize )
                    AddSample(
                              index, key, grad, ori,
                              irow + i, icol + j, rpos, cpos, rx, cx,par);
            }
        }
    }
    
    
    /* Given a sample from the image gradient, place it in the index array.
     */
    void AddSample(
                   float index[IndexSize][IndexSize][OriSize], keypoint& key,
                   const flimage& grad, const flimage& orim,
                   float r, float c, float rpos, float cpos, float rx, float cx,siftPar &par)
    {
        /* Clip at image boundaries. */
        if (r < 0  ||  r >= grad.nheight()  ||  c < 0  ||  c >= grad.nwidth())
            return;
        
        /* Compute Gaussian weight for sample, as function of radial distance
         from center.  Sigma is relative to half-width of index. */
        float	sigma  = par.IndexSigma * 0.5 * IndexSize,
		weight = exp(- (rpos * rpos + cpos * cpos) / (2.0 * sigma * sigma)),
        //		mag    = weight *  grad(c,r); 
		mag    = weight *  grad((int)c,(int)r); // Guoshen Yu, explicitely cast to int to avoid warning
        
        
        /* Subtract keypoint orientation to give ori relative to keypoint. */
        //	float	ori = orim(c,r) -  key.angle; 
        float	ori = orim((int)c,(int)r) -  key.angle; // Guoshen Yu, explicitely cast to int to avoid warning
        
        
        /* Put orientation in range [0, 2*PI].  If sign of gradient is to
         be ignored, then put in range [0, PI]. */
        if (par.IgnoreGradSign) {
            while (ori > PI ) ori -= PI;
            while (ori < 0.0) ori += PI;
        } else {
            while (ori > 2.0*PI) ori -= 2.0*PI;
            while (ori < 0.0   ) ori += 2.0*PI;
        }
        PlaceInIndex(index, mag, ori, rx, cx,par);
    }
    
    
    /* Increment the appropriate locations in the index to incorporate
     this image sample.  The location of the sample in the index is (rx,cx).
     */
    void PlaceInIndex(
                      float index[IndexSize][IndexSize][OriSize],
                      float mag, float ori, float rx, float cx,siftPar &par)
    {
        int	orr, rindex, cindex, oindex;
        float	rweight, cweight, oweight;
        float  *ivec;
        
        float	oval = OriSize * ori / (par.IgnoreGradSign ? PI : 2.0*PI);
        
        //	int	ri = (rx >= 0.0) ? rx : rx - 1.0,	/* Round down to next integer. */
        //		ci = (cx >= 0.0) ? cx : cx - 1.0,
        //		oi = (oval >= 0.0) ? oval : oval - 1.0;
        
        int	ri = (int)((rx >= 0.0) ? rx : rx - 1.0),	/* Round down to next integer. */ // Guoshen Yu, explicitely cast to int to avoid warning
		ci = (int)((cx >= 0.0) ? cx : cx - 1.0), // Guoshen Yu, explicitely cast to int to avoid warning
		oi = (int)((oval >= 0.0) ? oval : oval - 1.0); // Guoshen Yu, explicitely cast to int to avoid warning
        
        float	rfrac = rx - ri,			/* Fractional part of location. */
		cfrac = cx - ci,
		ofrac = oval - oi;
        assert(
               ri >= -1  &&  ri < IndexSize  &&
               oi >=  0  &&  oi <= OriSize  &&
               rfrac >= 0.0  &&  rfrac <= 1.0);
        
        /* Put appropriate fraction in each of 8 buckets around this point
         in the (row,col,ori) dimensions.  This loop is written for
         efficiency, as it is the inner loop of key sampling. */
        for (int r = 0; r < 2; r++) {
            rindex = ri + r;
            if (rindex >=0 && rindex < IndexSize) {
                rweight = mag * ((r == 0) ? 1.0 - rfrac : rfrac);
                
                for (int c = 0; c < 2; c++) {
                    cindex = ci + c;
                    if (cindex >=0 && cindex < IndexSize) {
                        cweight = rweight * ((c == 0) ? 1.0 - cfrac : cfrac);
                        ivec = index[rindex][cindex];
                        for (orr = 0; orr < 2; orr++) {
                            oindex = oi + orr;
                            if (oindex >= OriSize)  /* Orientation wraps around at PI. */
                                oindex = 0;
                            oweight = cweight * ((orr == 0) ? 1.0 - ofrac : ofrac);
                            ivec[oindex] += oweight;
                        }
                    }
                }
            }
        }
    }
    
    
    
    
    
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /// SIFT keypoint matching 
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    float DistSquared(keypoint &k1,keypoint &k2, float tdist, siftPar &par)
    {
        
        float dif;
        float distsq = 0.0;
        
        //	if (abs(k1.x - k2.x) > par.MatchXradius || abs(k1.y - k2.y) > par.MatchYradius) return tdist;
        
        if (ABS(k1.x - k2.x) > par.MatchXradius || ABS(k1.y - k2.y) > par.MatchYradius) return tdist;
        
        
        float *ik1 = k1.vec;
        float *ik2 = k2.vec;
        
        for (int i = 0; i < VecLength && distsq <= tdist; i++) {
            //    for (int i = 0; i < VecLength ; i++) {
            dif = ik1[i] - ik2[i];
            distsq += dif * dif;
            //distsq += ABS(dif);
            
            //		distsq += ((ik1[i] > ik2[i]) ? (ik1[i] - ik2[i]) : (-ik1[i] + ik2[i]));
            
        }
        
        return distsq;
    }
    
    float DistSquared_short(keypoint_short &k1,keypoint_short &k2, float tdist, siftPar &par)
    {
        // For Mac/Linux compilation using make: vectorization is possible with short.
        unsigned short distsq = 0;
        
        // For Windows compilation using Intel C++ compiler: vectorization is possible with int. 
        // int distsq = 0;
        
        if (ABS(k1.x - k2.x) > par.MatchXradius || ABS(k1.y - k2.y) > par.MatchYradius) return tdist;
        
        
        unsigned short *ik1 = k1.vec;
        unsigned short *ik2 = k2.vec;
        
        for (int i = 0; i < VecLength ; i++) {
            distsq += ((ik1[i] > ik2[i]) ? (ik1[i] - ik2[i]) : (-ik1[i] + ik2[i]));	
        }
        
        return distsq;
    }
    
    
    
    /* This searches through the keypoints in klist for the two closest
     matches to key.  It returns the ratio of the distance to key of the
     closest and next to closest keypoints in klist, while bestindex is the index
     of the closest keypoint.
     */
    float CheckForMatch(
                        keypoint& key, keypointslist& klist, int& min,siftPar &par)
    {	
        int	nexttomin = -1;
        float	dsq, distsq1, distsq2;
        distsq1 = distsq2 = 1000000000000.0f;
        
        for (int j=0; j< (int) klist.size(); j++){
            
            dsq = DistSquared(key, klist[j], distsq2,par);
            
            if (dsq < distsq1) {
                distsq2 = distsq1;
                distsq1 = dsq;
                nexttomin = min;
                min = j;
            } else if (dsq < distsq2) {
                distsq2 = dsq;
                nexttomin = j;
            }
        }
        
        return distsq1/distsq2 ;
    }
    
    float CheckForMatch_short(
                              keypoint_short& key, keypointslist_short& klist, int& min,siftPar &par)
    {	
        int	nexttomin = -1;
        float	dsq, distsq1, distsq2;
        distsq1 = distsq2 = 1000000000000.0f;
        
        for (int j=0; j< (int) klist.size(); j++){
            
            dsq = DistSquared_short(key, klist[j], distsq2,par);
            
            if (dsq < distsq1) {
                distsq2 = distsq1;
                distsq1 = dsq;
                nexttomin = min;
                min = j;
            } else if (dsq < distsq2) {
                distsq2 = dsq;
                nexttomin = j;
            }
        }
        
        return distsq1/distsq2 ;
    }
    
    
    
    void compute_sift_matches(
                              keypointslist& keys1,  keypointslist& keys2,
                              matchingslist& matchings,siftPar &par)
    {
        int	imatch=0;
        float	sqminratio = par.MatchRatio * par.MatchRatio,
		sqratio;
		
        // write the keypoint descriptors in char
        keypointslist_short keys1_short(keys1.size());
        for (int i=0; i< (int) keys1.size(); i++)
        {
            keys1_short[i].x = keys1[i].x;
            keys1_short[i].y = keys1[i].y;
            keys1_short[i].scale = keys1[i].scale;
            keys1_short[i].angle = keys1[i].angle;
            
            for (int k=0; k < VecLength; k++)
            {
                keys1_short[i].vec[k] = (unsigned short) (keys1[i].vec[k]);
            }
            
        }
        
        keypointslist_short keys2_short(keys2.size());
        for (int i=0; i< (int) keys2.size(); i++)
        {
            keys2_short[i].x = keys2[i].x;
            keys2_short[i].y = keys2[i].y;
            keys2_short[i].scale = keys2[i].scale;
            keys2_short[i].angle = keys2[i].angle;
            
            for (int k=0; k < VecLength; k++)
            {
                keys2_short[i].vec[k] = (unsigned short) (keys2[i].vec[k]);
            }
            
        }
        
        for (int i=0; i< (int) keys1.size(); i++) {
            
            //	sqratio = CheckForMatch(keys1[i], keys2, imatch,par);
            
            sqratio = CheckForMatch_short(keys1_short[i], keys2_short, imatch,par);
            
            
            if (sqratio< sqminratio)
                matchings.push_back( matching(keys1[i],keys2[imatch] ));
            //		 matchings.push_back( matching_char(keys1_char[i],keys2_char[imatch] ));
        }
    }
    
    
}