// Authors: Unknown. Please, if you are the author of this file, or if you 
// know who are the authors of this file, let us know, so we can give the 
// adequate credits and/or get the adequate authorizations.

// WARNING: 
// This file implements an algorithm possibly linked to the patent
//
// David Lowe  "Method and apparatus for identifying scale invariant 
// features in an image and use of same for locating an object in an 
// image",  U.S. Patent 6,711,293.
//
// This file is made available for the exclusive aim of serving as
// scientific tool to verify of the soundness and
// completeness of the algorithm description. Compilation,
// execution and redistribution of this file may violate exclusive
// patents rights in certain countries.
// The situation being different for every country and changing
// over time, it is your responsibility to determine which patent
// rights restrictions apply to you before you compile, use,
// modify, or redistribute this file. A patent lawyer is qualified
// to make this determination.
// If and only if they don't conflict with any patent terms, you
// can benefit from the following license terms attached to this
// file.
//
// This program is provided for scientific and educational only:
// you can use and/or modify it for these purposes, but you are
// not allowed to redistribute this work or derivative works in
// source or executable form. A license must be obtained from the
// patent right holders for any other use.


#ifndef _LIBSIFT_IPOL_H_
#define _LIBSIFT_IPOL_H_


#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <fstream>
#include <cmath>
#include <cassert>
#include <vector>
#include <getopt.h>



namespace libSiftIPOLCODE
{
    


    
    
    
    
#define MAX(i,j) ( (i)<(j) ? (j):(i) )
#define MIN(i,j) ( (i)<(j) ? (i):(j) )
    
#ifndef LUTMAX
#define LUTMAX 30
#endif
#define LUTPRECISION 1000.0
    
#define TINY 1.0e-10
    //#define MAXFLOAT 10000000.0
    
    
#define IRAC8    0.35355339  /* 1/sqrt(8)   */
#define IRAC2P2  0.29289322  /* 1/(sqrt(2)+2) */
#define IRAC2    0.70710678  /* 1/sqrt(2) */
#define RAC8P4   6.8284271   /* sqrt(8)+4 */
#define RADIANS_TO_DEGREES (180.0/M_PI)
 
#ifndef PI   
#define PI 3.14159
#endif    
    
#define COEFF_YR 0.299
#define COEFF_YG 0.587
#define COEFF_YB 0.114
    
    
    
    
    ///////////////////////////////////////////////////////////////// Errors
    void wxwarning(const char * message,const char *function, const char *file);
    void wxerror(const char * message, const char *function, const char *file);
    
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////   Used and checked mathematical functions
    double fsqr(double a);
    
    void  fill_exp_lut(float *lut,int size); /* Fills exp(x) for x great or equal than zero*/
    float slut(float dif,float *lut); /* We look for f(dif) in the lut*/
    
    
    
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////   Used and checked mxsignal functions
    float max(float *u,int *pos, int size);    /// Max(u), pos contains the index of the maximum 
    float min(float *u,int *pos, int size);    /// Min(u), pos contains the index of the minimum 	
    void max_u_v(float *u,float *v,int size); 
    void max_u_k(float *u,float k,int size);
    void min_u_v(float *u,float *v,int size);
    void min_u_k(float *u,float k,int size);
    
    
    
    //void abs(float *u,float *v,int size);	 	/// v = abs(u)
    
    void  copy(float *u,float *v,int size);			/// v = u
    void  clear(float *u, float value ,int size);        	/// u = k 
    void  combine(float *u,float a,float *v,float b, float *w,  int size);    /// w = a*u + b*v 
    void  multiple(float *u,float multiplier,int size);         	/// u = K * u  
    
    
    float scalar_product(float *u, float *v, int n);
    
    // float lpdist(float *u,float *v,float *mask,int pow,int size);   
    // float lpnorm(float *u,int fpow,int size);   
    
    float mean(float *u,int size);
    float var(float *u,int size);
    float median(float *u,int size);
    
    float nearest(float *u,float value,int *pos,int size);  /// Returns the nearest value in the vector u and its position if selected
    void binarize(float *u, float *v,float value, int inverse, int size);    /// v = 255 if u > value 0 else  
    int normalize(float *u,int size);    			/// u = u / sum_i u(i).   Returns 0 if mxsignal sum equals zero 
    
    
    
    float *  gauss(int sflag,float std,int *size);  	/// Create a 1d gauss kernel of standard deviation std
    
    // void addnoise(float *u,float *v,float std,long int randinit, int size);    
    // void addnoise_var_afine(float *u,float *v,float a,float b,long int randinit, int size);    
    
    
    void quick_sort(float *arr,float *brr,int n);  		/// Quicksort
    
    
    /// histogram of values. 'n' (number of bins) or  's' (step) must be selected in flag while the other value is filled 
    float * histo ( float* input, float *iminim, float *imaxim, int *n, float *s, int size, char flag );	
    
    
    
    
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////   Used and checked image functions
    
    void compute_gradient_orientation(float* igray,float *grad, float *ori, int width, int height);
    
    // void extract ( float *igray,float *ogray, int ax, int ay,int cwidth, int cweight,int width, int height );
    
    void sample ( float *igray,float *ogray, float factor ,int width, int height);
    void sample_aglomeration(float *igray,float *ogray, float factor, int width, int height);
    
    void gray ( float *red, float *green,float *blue, float *out, int width, int height );
    
    /*
     float l2_distance ( float * u0,float *u1,int i0,int j0,int i1,int j1,int radius,int width,int height);
     float l2_distance_non_normalized(float *u0,float *u1,int i0,int j0,int i1,int j1,int radius,int width,int height);
     float weighted_l2_distance ( float *u0,float *u1,int i0,int j0,int i1,int j1,int width,int height,float * kernel,int radius );
     float l2_distance_nsq(float *u0,float *u1,int i0,int j0,int i1,int j1,int xradius, int yradius,int width,int height);
     float weighted_l2_distance_nsq(float *u0,float *u1,int i0,int j0,int i1,int j1,int width,int height,float *kernel,int xradius, int yradius);
     */
    
    void rgb2yuv ( float *r,float *g,float *b,float *y,float *u,float *v,int width,int height );  
    void yuv2rgb ( float *r,float *g,float *b,float *y,float *u,float *v,int width,int height ); 
    
    void rgb2yuv(float *r,float *g,float *b,float *y,float *u,float *v,float yR, float yG, float yB, int width,int height);
    void yuv2rgb(float *r,float *g,float *b,float *y,float *u,float *v,float yR, float yG, float yB, int width,int height);
    
    
    void draw_line(float *igray, int a0, int b0, int a1, int b1, float value, int width, int height);
    // void draw_circle(float *igray, int pi,int pj,float radius, float value, int width, int height);
    void draw_square(float *igray, int a0, int b0, int w0, int h0, float value, int width, int height);
    
    
    
    
    
    float * directional_gauss_filter(float xsigma, float ysigma, float angle, int *kwidth, int *kheight);
    
    
    void median(float *u,float *v, float radius, int niter, int width,int height);
   
    
    /// Convolution with a  separable kernel,  boundary condition: 0=zero, 1=symmetry
    void separable_convolution(float *u, float *v, int width, int height, float *xkernel, int xsize, float *ykernel, int ysize,int boundary);     
    
    void buffer_convolution(float *buffer,float *kernel,int size,int ksize);
    void horizontal_convolution(float *u, float *v, int width, int height, float *kernel, int ksize, int boundary);
    void vertical_convolution(float *u, float *v, int width, int height, float *kernel,int ksize, int boundary);
    
    void fast_separable_convolution(float *u, float *v, int width, int height,float * xkernel, int xsize,float *ykernel,int ysize,int boundary);
    
    /// Can be called with u=v
    void gaussian_convolution(float *u, float *v, int width, int height, float sigma);
    void gaussian_convolution(float *u, float *v, int width, int height, float sigma, int ksize);
    
    void convol(float *u, float *v, int width, int height, float *kernel, int kwidth, int kheight);    /// Convolution with a kernel,  No padding applied to the image
    
    void heat(float *u, float *v, float step, int niter, float sigma, int width, int height);
    
    
    
    
    
    
#define NRMAX(i,j) ( (i)<(j) ? (j):(i) )
#define NRMIN(i,j) ( (i)<(j) ? (i):(j) )
#define NRTINY 1.0e-10
    
    
    // **********************************************
    //  float **  basic functions
    // **********************************************
    
    float ** allocate_float_matrix(int nrows, int ncols);
    
    void desallocate_float_matrix(float **matrix, int nrows, int ncols);
    
    // **********************************************
    //  LU based algorithms
    // **********************************************
    
    // Solves Ax=b by using lu decomposition 
    // a matrix a[1..n][1..n] is replaced by the LU decompositions of a rowwise permutation of itself
    // b[1..n] and x[1..n]
    int lusolve(float **a, float *x, float *b, int n);
    
    
    /*-- LU decomposition */
    /* Given a matrix a[1..n][1..n] this routine replacess it by the LU decompositions of a rowwise permutation of itself. */
    /* a and n are input, a is output, arranged as in equation (2.3.14) above; indx[1..n] in an output vector that records */
    /* the row permutation effected by the partial pivoting; d is output as +-1 depending on whether the number of row     */
    /* interchanges was even or odd respectively.                                                                          */
    int ludcmp(float **a, int n, int *indx, float *d); /* LU decomposition */
    
    /* Solves the set of n linear equations Ax=b. Here a[0..n-1][0..n-1] as input, not as the matrix A but rather as its LU decomposition,*/
    /* determined by the routine ludcmp. indx[0..n-1] is input as the permutation vector returned by ludcmp. b[0..n-1] is input as the    */
    /* right hand side vector and returns with the solution vector x. */
    void lubksb(float **a, int n, int *indx, float *b); /* LU linear solution */
    
    
    
    
    
    
    
    class flimage {
        
    private:
        
        int	width, height;	// image size
        float*	p;		// array of color levels: level of pixel (x,y) is p[y*width+x]
        
    public:
        
        
        //// Construction
        flimage();
        flimage(int w, int h);
        flimage(int w, int h, float v);
        flimage(int w, int h, float* v);
        flimage(const flimage& im);
        flimage& operator= (const flimage& im);
        
        
        void create(int w, int h);
        void create(int w, int h, float *v);
        
        //// Destruction
        void erase();
        ~flimage();
        
        //// Get Basic Data	
        int nwidth() const {return width;} 	// image size
        int nheight() const {return height;} 
        
        /// Access values
        float* getPlane() {return p;}	// return the adress of the array of values 
        
        float operator()(int x, int y) const {return p[ y*width + x ];} 	// acces to the (x,y) value
        float& operator()(int x, int y) {return p[ y*width + x ];}	// by value (for const images) and by reference
        
        
    };
    
    
    
    
    
    ///////////// Description
    /// For each octave:
    ///    	- Divide in par.Scales scales
    /// 	- Convolve and compute differences of convolved scales
    ///	- Look for a 3x3 multiscale extrema and contraste enough and with no predominant direction (no 1d edge)
    
    /// For each extrema
    ///	- Compute orientation histogram in neighborhood.
    ///	- Generate a keypoint for each mode with this orientation
    
    
    /// For each keypoint
    ///	 - Create vector 
    
    
    
    ///////////// Possible differences with MW
    /// Gaussian convolution
    
    

    
    
    // BASIC STRUCTURES:
    
    // Keypoints:
#define OriSize  8
#define IndexSize  4
#define VecLength  IndexSize * IndexSize * OriSize
    
    
    /* Keypoint structure:
     position:	x,y
     scale:		s
     orientation:	angle
     descriptor:	array of gradient orientation histograms in a neighbors */
    struct keypoint {
        float	x,y,
		scale,
		angle;
        float	vec[VecLength];
    };
    
    
    /* Keypoint structure:
     position:	x,y
     scale:		s
     orientation:	angle
     descriptor:	array of gradient orientation histograms in a neighbors */
    struct keypoint_char {
        float	x,y,
        scale,
        angle;
        unsigned char	vec[VecLength];
    };
    
    /* Keypoint structure:
     position:	x,y
     scale:		s
     orientation:	angle
     descriptor:	array of gradient orientation histograms in a neighbors */
    struct keypoint_short {
        float	x,y,
        scale,
        angle;
        unsigned short	vec[VecLength];
    };
    
    /* Keypoint structure:
     position:	x,y
     scale:		s
     orientation:	angle
     descriptor:	array of gradient orientation histograms in a neighbors */
    struct keypoint_int {
        float	x,y,
        scale,
        angle;
        unsigned int	vec[VecLength];
    };
    
    /* List of keypoints: just use the standard class vector: */
    typedef std::vector<keypoint> keypointslist;
    
    /* List of keypoints: just use the standard class vector: */
    typedef std::vector<keypoint_char> keypointslist_char;
    typedef std::vector<keypoint_short> keypointslist_short;
    typedef std::vector<keypoint_int> keypointslist_int;
    
    
    
    /* Matching: just use the standard class pair: */
    typedef std::pair<keypoint,keypoint> matching;
    
    
    /* List of matchings: just use the standard class vector: */
    typedef std::vector<matching> matchingslist;
    
    
    struct siftPar
    {
        
        int OctaveMax;
        
        int DoubleImSize;
        
        int order;
        
        
        /* InitSigma gives the amount of smoothing applied to the image at the
         first level of each octave.  In effect, this determines the sampling
         needed in the image domain relative to amount of smoothing.  Good
         values determined experimentally are in the range 1.2 to 1.8.
         */
        float  InitSigma /*= 1.6*/;
        
        
        /* Peaks in the DOG function must be at least BorderDist samples away
         from the image border, at whatever sampling is used for that scale.
         Keypoints close to the border (BorderDist < about 15) will have part
         of the descriptor landing outside the image, which is approximated by
         having the closest image pixel replicated.  However, to perform as much
         matching as possible close to the edge, use BorderDist of 4.
         */
        int BorderDist /*= 5*/;
        
        
        /* Scales gives the number of discrete smoothing levels within each octave.
         For example, Scales = 2 implies dividing octave into 2 intervals, so
         smoothing for each scale sample is sqrt(2) more than previous level.
         Value of 2 works well, but higher values find somewhat more keypoints.
         */
        
        int Scales /*= 3*/;
        
        
        /// Decreasing PeakThresh allows more non contrasted keypoints
        /* Magnitude of difference-of-Gaussian value at a keypoint must be above
         this threshold.  This avoids considering points with very low contrast
         that are dominated by noise.  It is divided by Scales because more
         closely spaced scale samples produce smaller DOG values.  A value of
         0.08 considers only the most stable keypoints, but applications may
         wish to use lower values such as 0.02 to find keypoints from low-contast
         regions.
         */
        
        //#define  PeakThreshInit  255*0.04 
        //#define  PeakThresh      PeakThreshInit / Scales
        float PeakThresh  /*255.0 * 0.04 / 3.0*/;
        
        /// Decreasing EdgeThresh allows more edge points
        /* This threshold eliminates responses at edges.  A value of 0.08 means
         that the ratio of the largest to smallest eigenvalues (principle
         curvatures) is below 10.  A value of 0.14 means ratio is less than 5.
         A value of 0.0 does not eliminate any responses.
         Threshold at first octave is different.
         */
        float  EdgeThresh  /*0.06*/;
        float  EdgeThresh1 /*0.08*/;
        
        
        /* OriBins gives the number of bins in the histogram (36 gives 10
         degree spacing of bins).
         */
        int OriBins  /*36*/;
        
        
        /* Size of Gaussian used to select orientations as multiple of scale
         of smaller Gaussian in DOG function used to find keypoint.
         Best values: 1.0 for UseHistogramOri = FALSE; 1.5 for TRUE.
         */
        float OriSigma  /*1.5*/;
        
        
        /// Look for local (3-neighborhood) maximum with valuer larger or equal than OriHistThresh * maxval
        ///  Setting one returns a single peak
        /* All local peaks in the orientation histogram are used to generate
         keypoints, as long as the local peak is within OriHistThresh of
         the maximum peak.  A value of 1.0 only selects a single orientation
         at each location.
         */
        float OriHistThresh  /*0.8*/;
        
        
        /// Feature vector is normalized to has euclidean norm 1.
        /// This threshold avoid the excessive concentration of information on single peaks
        /* Index values are thresholded at this value so that regions with
         high gradients do not need to match precisely in magnitude.
         Best value should be determined experimentally.  Value of 1.0
         has no effect.  Value of 0.2 is significantly better.
         */
        float  MaxIndexVal  /*0.2*/;
        
        
        /* This constant specifies how large a region is covered by each index
         vector bin.  It gives the spacing of index samples in terms of
         pixels at this scale (which is then multiplied by the scale of a
         keypoint).  It should be set experimentally to as small a value as
         possible to keep features local (good values are in range 3 to 5).
         */
        int  MagFactor   /*3*/;
        
        
        /* Width of Gaussian weighting window for index vector values.  It is
         given relative to half-width of index, so value of 1.0 means that
         weight has fallen to about half near corners of index patch.  A
         value of 1.0 works slightly better than large values (which are
         equivalent to not using weighting).  Value of 0.5 is considerably
         worse.
         */
        float   IndexSigma  /*1.0*/;
        
        /* If this is TRUE, then treat gradients with opposite signs as being
         the same.  In theory, this could create more illumination invariance,
         but generally harms performance in practice.
         */
        int  IgnoreGradSign  /*0*/;
        
        
        
        float MatchRatio  /*0.6*/;
        
        /*
         In order to constrain the research zone for matches.
         Useful for example when looking only at epipolar lines
         */
        
        float MatchXradius /*= 1000000.0f*/;
        float MatchYradius /*= 1000000.0f*/;
        
        int noncorrectlylocalized;
        
    };
    
    //////////////////////////////////////////////////////////
    /// SIFT 
    //////////////////////////////////////////////////////////
    
    void default_sift_parameters(siftPar &par);
    
    void compute_sift_keypoints(float *input,  keypointslist& keypoints,int width, int height, siftPar &par);
    
    
    
    
    // MATCHING DETECTION FUNCTION:
    void compute_sift_matches( keypointslist& keys1, keypointslist& keys2, matchingslist& matchings, siftPar &par);
    
    
}

#endif // _LIBSIFT_H_



